#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>

#include <errno.h>

#include "sqirc.h"

struct connection **connections;
unsigned long connection_count = 0;
unsigned long client_command_count = 0;
unsigned long server_command_count = 0;
unsigned long ctcp_command_count = 0;

static void (*sqirc_connection_cleanup)(struct connection *) = 0;
static void (*sqirc_connection_read_data)(struct connection *, gint) = 0;
static void (*sqirc_error_handler)(struct connection *, ulong, int, char *, struct callback_data *) = 0;

static int sqirc_send_queue(struct connection *);
static void sqirc_internal_read_data(struct connection *, gint);
static void sqirc_internal_cleanup(struct connection *);
static SQIRC_FUNC_ERRORHANDLER(sqirc_internal_error_handler);

static struct server_command **server_commands = 0;
static struct client_command **client_commands = 0;
static struct client_command **ctcp_commands = 0;

/*
 * to provide more convenience to the developer, we'll
 * handle some commands first, and then give away control
 * to the developers callback
 * PRIVMSG is needed for direct CTCP handling
 */
static SQIRC_FUNC_CMDCLIENT(internal_clientcmd_privmsg);
static SQIRC_FUNC_CMDCLIENT(internal_clientcmd_ping);
static struct client_command internal_client_commands[] =
{ { "PRIVMSG",           internal_clientcmd_privmsg },
	{ "NOTICE",            internal_clientcmd_privmsg },
	{ "PING",              internal_clientcmd_ping },
	LAST_COMMAND
};

static SQIRC_FUNC_CMDCLIENT(internal_ctcpcmd_ping);
static struct client_command internal_ctcp_commands[] =
{ { "PING",              internal_ctcpcmd_ping },
	LAST_COMMAND
};

static SQIRC_FUNC_CMDSERVER(internal_servercmd_motd);
static struct server_command internal_server_commands[] =
{
	{ RPL_MOTDSTART,       internal_servercmd_motd },
	{ RPL_MOTD,            internal_servercmd_motd },
	{ RPL_ENDOFMOTD,       internal_servercmd_motd },
	{ ERR_NOMOTD,          internal_servercmd_motd },
	{ 0, NULL }
};

void sqirc_initialize(struct server_command *s_cmd,
											struct client_command *cl_cmd,
											struct client_command *ct_cmd)
{
	/* register internal commands structures */
	sqirc_register_client_commands(internal_client_commands);
	sqirc_register_server_commands(internal_server_commands);
	sqirc_register_ctcp_commands(internal_ctcp_commands);
	
	/* register developer provided commandstructures */
	sqirc_register_server_commands(s_cmd);
	sqirc_register_client_commands(cl_cmd);
	sqirc_register_ctcp_commands(ct_cmd);
	
	sqirc_connection_cleanup = sqirc_internal_cleanup;
	sqirc_connection_read_data = sqirc_internal_read_data;
	sqirc_error_handler = sqirc_internal_error_handler;
	
}

void sqirc_register_client_commands(struct client_command *commands)
{
	int num = client_command_count + 1;

	if (commands == 0)
		return;

	/* allocate memory block */
	client_commands = realloc(client_commands, sizeof(void*)*(num + 1));
	
	/* append new options */
	client_commands[client_command_count]   = commands;
	client_commands[client_command_count+1] = 0;
	
	client_command_count++;
}

void sqirc_register_server_commands(struct server_command *commands)
{
	int num = server_command_count + 1;

	if (commands == 0)
		return;
	
	/* allocate memory block */
	server_commands = realloc(server_commands, sizeof(void*)*(num + 1));

	/* append new options */
	server_commands[server_command_count]   = commands;
	server_commands[server_command_count+1] = 0;

	server_command_count++;
}

void sqirc_register_ctcp_commands(struct client_command *commands)
{	
	int num = ctcp_command_count + 1;

	if (commands == 0)
		return;

	/* allocate memory block */
	ctcp_commands = realloc(ctcp_commands, sizeof(void*)*(num + 1));

	/* append new options */
	ctcp_commands[ctcp_command_count]   = commands;
	ctcp_commands[ctcp_command_count+1] = 0;
	
	ctcp_command_count++;
}

void sqirc_register_callback(int type, void (*callback)())
{
	switch(type)
		{
		case IRC_FUNC_CLEANUP:
			sqirc_connection_cleanup = callback;
			break;
		case IRC_FUNC_READDATA:
			sqirc_connection_read_data = callback;
			break;
		case IRC_FUNC_ERRORHANDLER:
			sqirc_error_handler = callback;
			break;
		default:
			break;
		}
}

/* create a new connection stucture, register it with connections
 * and bzero it's structure
 */
struct connection *sqirc_new_connection()
{
	int num;

	/* try to find an unused connection, or ... */
	if (connection_count > 0)
		for (num = 0; connections[num]; num++)
			if (connections[num]->status == IRC_UNUSED)
				{
					connections[num]->status = IRC_UNCONNECTED;
					return connections[num];
				}

	/* ... create a new one */
	num = connection_count + 1;
	connections = realloc(connections, sizeof(void*) * (num + 1));
	
	connections[connection_count] = malloc(sizeof(struct connection));

	bzero(connections[connection_count], sizeof(struct connection));
	connections[connection_count]->socket = IRC_UNUSED;
	connections[connection_count]->status = IRC_UNUSED;
	connections[connection_count]->input_id = -1;
	connections[connection_count]->queue = 0;
	
	connections[connection_count+1] = 0;
	return connections[connection_count++];
}

void sqirc_close_connection(struct connection *connection)
{
	if (connection->input_id != -1)
		gdk_input_remove(connection->input_id);
	
	/* call the application's connection cleanup if data present */
	if (sqirc_connection_cleanup != 0)
		sqirc_connection_cleanup(connection);
	
	connection->data = 0;
	close(connection->socket);
	connection->socket = connection->status = IRC_UNUSED;
	
}

void sqirc_close_all_connections()
{
	int num;
	
	if (connections == 0)
		return;
	
	for (num = 0; connections[num]; num++)
		{
			if (connections[num]->status != IRC_UNUSED)
				sqirc_close_connection(connections[num]);
			free(connections[num]);
			connections[num] = 0;
		}
	
	free(connections);
	free(client_commands);
	free(server_commands);
	free(ctcp_commands);

}

int sqirc_send(struct connection *connection, char *fmt, ...)
{
	va_list args;
	char net_buffer[IRC_BUFSIZE];
	time_t now = time(0);
	
	if (!connection)
		{
			sqirc_error_handler(0, IRCERR_NOCONNECTION, 0, "no connection", 0);
			return -1;
		}
	if (connection->status != IRC_CONNECTED)
		{
			sqirc_error_handler(0, IRCERR_NOTCONNECTED, 0, "not connected", 0);
			return -1;
		}
	
	va_start(args, fmt);
	vsnprintf(net_buffer, IRC_BUFSIZE*4, fmt, args);
	va_end(args);

	if (connection->last_send != now)
		{
			connection->last_send = now;
			connection->bytes_sent = 0;
		}
	else
		connection->bytes_sent += strlen(net_buffer);
	
	if (connection->bytes_sent > IRC_SENDMAX
			|| connection->queue)
		{
			char *message;
			
			if ( (message = strdup(net_buffer)) == NULL)
				{
					perror("strdup");
					exit(1);
				}
			if (!connection->queue)
				gtk_timeout_add(IRC_QUEUETIMEOUT,
												(GtkFunction)sqirc_send_queue,
												connection);
			IRC_ADD_LIST_ITEM(connection->queue, message);
			return 1;
		}
	
	if (connection->flags & SQIRC_DEBUG)
		sqirc_error_handler(connection, IRCDEBUG_NET_OUT, 0, net_buffer, 0);
	
	return send(connection->socket,	net_buffer,	strlen(net_buffer), 0);
}

/*
 * search the client_commands list, and return the found data in c_cmd
 * start tells at what index to start searching
 */
int sqirc_find_client_command(const char *name, int start,
															struct client_command *c_cmd)
{
	int i=0, c=0;
	
	bzero(c_cmd, sizeof(struct client_command));
	for (c = start; client_commands[c]; c++)
		for (i = 0; client_commands[c][i].command[0]; i++)
			{
				if (client_commands[c][i].callback
						&& !strcmp(client_commands[c][i].command, name))
					{
						*c_cmd = client_commands[c][i];
						return 1;
					}
			}
	return 0;
}

/*
 * search the ctcp_commands list, and return the found data in c_cmd
 * start tells at what index to start searching
 */
int sqirc_find_ctcp_command(const char *name, int start,
														struct client_command *c_cmd)
{
	int i, c;
	
	for (c = start; ctcp_commands[c]; c++)
		for (i=0; ctcp_commands[c][i].command[0]; i++)
			{
				if (!strcmp(ctcp_commands[c][i].command, name))
					{
						*c_cmd = ctcp_commands[c][i];
						return 1;
					}
			}
	return 0;
}

/*
 * search the server_commands list, and return the found data in s_cmd
 * start tells at what index to start searching
 */
int sqirc_find_server_command(int reply, int start,
															 struct server_command *s_cmd)
{
	int i, c;

	bzero(s_cmd, sizeof(struct server_command));
	for (c = start; server_commands[c]; c++)
		for(i = 0; server_commands[c][i].reply; i++)
			{
				if (server_commands[c][i].reply == reply)
					{
						*s_cmd = server_commands[c][i];
						return 1;
					}
			}
	return 0;
}


int sqirc_connect(const char *hostname,
									const int port,
									const int flags,
									struct irc_userdata *userdata,
									struct connection **conn)
{
  /* use name to find TCP/IP address of server */
  struct sockaddr_in src;
  struct hostent *ph;
  long address;
  int sw;
	struct connection *connection = 0;
	
	/* error checking, just in case .... */
	if (!userdata->nickname)
		return IRCERR_DATAMISSING;
	if (!userdata->username)
		return IRCERR_DATAMISSING;
	if (!userdata->desc)
		return IRCERR_DATAMISSING;
	if (!hostname)
		return IRCERR_DATAMISSING;
	
  if( isdigit((size_t)*hostname) )
		{
			if( (address = inet_addr(hostname)) == -1 )
				return IRCERR_INVALID_HOSTNAME;
			src.sin_addr.s_addr = address;
			src.sin_family = AF_INET;
		}
  else
		{
			if ( (ph = gethostbyname(hostname)) == NULL )
				return IRCERR_LOOKUP_FAIL;
			src.sin_family = ph->h_addrtype;
			memcpy(&src.sin_addr, ph->h_addr, ph->h_length);
		}
  src.sin_port = htons(port);
  
  /* ignore broken pipes */
  signal (SIGPIPE, SIG_IGN);
	
  *conn = connection = sqirc_new_connection();
	connection->flags = flags;
	memcpy(&connection->userdata, userdata, sizeof(struct irc_userdata));
  if( (connection->socket = socket (AF_INET, SOCK_STREAM, 0)) < 0 )
		return IRCERR_NOSOCKET;
  
  sw = 1;
  setsockopt (connection->socket,SOL_SOCKET,SO_LINGER,
							(char *)&sw,sizeof(sw));
  sw = 1;
  setsockopt (connection->socket,SOL_SOCKET,SO_REUSEADDR,
							(char *)&sw,sizeof(sw));
  sw = 1;
  setsockopt (connection->socket,SOL_SOCKET,SO_KEEPALIVE,
							(char *)&sw,sizeof(sw));
  
  if( connect(connection->socket, (struct sockaddr *) &src, sizeof (src)) < 0)
		{
			sqirc_close_connection(connection);
			return IRCERR_NOCONNECTION;
		}
	
	if (!sqirc_connection_read_data)
		sqirc_connection_read_data = sqirc_internal_read_data;
  connection->status = IRC_CONNECTED;
  connection->input_id
		= gdk_input_add(connection->socket,
										GDK_INPUT_READ|GDK_INPUT_EXCEPTION,
										(GdkInputFunction)sqirc_connection_read_data,
										(void*)connection);
  fcntl(connection->socket, F_SETFL, O_NONBLOCK);
  
  /* now send identify package */
	sqirc_send(connection, "NICK %s\r\n", userdata->nickname);
	sqirc_send(connection, "USER %s localhost %s :%s\r\n",
						 userdata->username, userdata->nickname, userdata->desc);
	
  return IRC_CONNECTED;
}

/*
 * parse_from gives you the parts of a from token in 'from' and returns 1
 * if it contains a nickname, or 0 if it only was a servername
 */
int sqirc_parse_from(char *token, struct from_data *from)
{
	char *cp1, *cp2;
	int ret = 0;
	
	cp1 = token;
	cp2 = from->nick;
	
	/* azzit!azzit@horizon: nickname!username@hostname */
	while ( *cp1 != 0 && *cp1 != '!' )
		*cp2++ = *cp1++;
	*cp2 = 0;
	
	/* it's a servername; no ! in it */
	if (*cp1 != '!')
		{
			ret = 0;
			strcpy(from->host, from->nick);
			from->nick[0] = 0;
			return ret;
		}
	
	/* read username part of the token buffer */
	cp1++;
	cp2 = from->user;
	while ( *cp1 != 0 && *cp1 != '@' )
		*cp2++ = *cp1++;
	*cp2 = 0;
	
	/* assume that the remainder is the hostname */
	strncpy(from->host, ++cp1, IRC_BUFSIZE);
	
	return ret;
}

int sqirc_process_buffer(struct connection *connection)
{
	char *cp1, *cp2;                   /* some char pointer */
	char *eos, *eob;                   /* end of string, end of buffer */
	int i, c;                          /* a counter or two: always useful */
	char name[IRC_BUFSIZE];            /* tmp storage of command name */
	char msg[IRC_BUFSIZE];             /* whole message 'body' */
	char word[IRC_BUFSIZE];            /* tmp buffer for a word in the list */
	int reply;
	struct callback_data cb_data;
	struct client_command c_cmd;
	struct server_command s_cmd;
	
	if (connection->flags & SQIRC_DEBUG)
		sqirc_error_handler(connection, IRCDEBUG_NET_IN, 0, connection->buffer, 0);
	
#define INIT_EOS eos = (cp2 + IRC_BUFSIZE + 1)   /* initialize end of string */
#define READ_STR while ( *cp1 != 0 && cp2 != eos && !isspace(*cp1)) \
                   *cp2++ = *cp1++
#define SKIP_SPACE while (*cp1 != 0 && isspace(*cp1)) cp1++
	
	/* initialization of needed data */
	bzero(name, IRC_BUFSIZE);
	bzero(msg, IRC_BUFSIZE);
	bzero(&cb_data, sizeof(cb_data));
	bzero(&s_cmd, sizeof(s_cmd));
	bzero(&c_cmd, sizeof(c_cmd));
	reply = 0;
	
	/* copy the buffer into the callback data */
	strncpy(cb_data.buffer, connection->buffer, 2048);
	cp1 = connection->buffer;
	eob = cp1 + strlen(connection->buffer);   /* initialize end of buffer */
	
	if (*cp1 == ':')
		{
			cp1++;                         /* skip colon */
			cp2 = cb_data.from.token;
			INIT_EOS;
			
			/* read string */
			READ_STR;
			*cp2 = 0;
			cp1++;                         /* skip whitespace */
			sqirc_parse_from(cb_data.from.token, &cb_data.from);
		}
	
	cp2 = name;
	INIT_EOS;
	READ_STR;
	SKIP_SPACE;
	
	if (*cp1 == '*')          /* hybrid sends: 433 * <nick> */
		cp1 += 2;
	
	if (*cp1 != ':')
		{
			cp2 = cb_data.nick;
			INIT_EOS;
			READ_STR;
			SKIP_SPACE;
		}
	
	strncpy(cb_data.message, cp1, IRC_BUFSIZE);
	
	cb_data.wcount = 0;
	/* initialize buffer */
	cb_data.list = calloc(64,	sizeof(char*));
	while (*cp1 != 0)
		{
			
			/* read the next word */
			bzero(word, IRC_BUFSIZE);            
			cp2 = word;
			INIT_EOS;
			READ_STR;
			SKIP_SPACE;

			/* assign word to callback_data */
			cb_data.list[cb_data.wcount] = strdup(word);
			cb_data.list[++cb_data.wcount] = 0;
		}
	
	if ( (reply = atoi(name)) )
		{
			/* find the callback for this reply */
			if (server_commands)
				sqirc_find_server_command(reply, 0, &s_cmd);
			
			/* check if the search was successful */
			if (!s_cmd.reply)
				sqirc_error_handler(connection, IRCERR_UNHANDLED_SERVER_CMD,
														reply, 0, &cb_data);
			else if (s_cmd.callback)  /* yeah, cool */
				s_cmd.callback(connection, reply, &cb_data);
		}
	else 
		{
			/* find the callback for this command */
			if ( client_commands && sqirc_find_client_command(name, 0, &c_cmd) != 0 )
				c_cmd.callback(connection, name, &cb_data);
			else
				sqirc_error_handler(connection, IRCERR_UNHANDLED_CLIENT_CMD,
														0, name, &cb_data);
			
		}
	
	for (i = 0; cb_data.list[i]; i++)
		free(cb_data.list[i]);
	free(cb_data.list);
	
	return 0;
}

int sqirc_send_queue(struct connection *connection)
{
  int len;
  char *buf = 0;
  time_t now = time(0);
	
  while(connection->queue)
		{
			buf = (char *)connection->queue->data;
			
			if (connection->bytes_sent > IRC_SENDMAX)
				return 1;                  /* don't remove the timeout handler */
			
			len = strlen(buf);
			if (connection->last_send != now)
				{
					connection->last_send = now;
					connection->bytes_sent = 0;
				}
			else
				connection->bytes_sent += len;
			
			if (connection->flags & SQIRC_DEBUG)
				sqirc_error_handler(connection, IRCDEBUG_NET_OUT, 0, buf, 0);
			
			send(connection->socket, buf, len, 0);
			
			IRC_DEL_LIST_ITEM(connection->queue, buf);
			free (buf);
		}
  return 0;       /* return, telling gdk to remove the timeout handler */
}

/* internal stuff to handler various situatations */
void sqirc_internal_read_data(struct connection *connection, gint sok)
{
  int len;
  char lbuf[2050];

  while (1)
		{
			len = recv (connection->socket, lbuf, sizeof lbuf - 2, 0);
			if (len < 1)
				{
					if (len < 0)
						{
							if (errno == EAGAIN || errno == EWOULDBLOCK)
								return;
						}
					sqirc_close_connection(connection);
					return;
				}
			else
				{
					int i = 0;
					lbuf[len] = 0;
		
					while (i < len)
						{
							switch (lbuf[i])
								{
								case '\r':
									break;
			
								case '\n':
									connection->buffer[connection->pos] = 0;
									sqirc_process_buffer(connection);
									connection->pos = 0;
									break;
									
								default:
									connection->buffer[connection->pos] = lbuf[i];
									connection->pos++;
									if (connection->pos == 2047)
										{
											sqirc_error_handler(connection, IRCERR_BUFFEROVERFLOW,
																					0, "Buffer overflow", 0);
											connection->pos = 2046;
										}
									break;
								}
							i++;
						}
				}
		}
}

/* nothing to do on internal cleanup */
void sqirc_internal_cleanup(struct connection *connection)
{
}

SQIRC_FUNC_ERRORHANDLER(sqirc_internal_error_handler)
{
	switch(irc_error)
		{
			/* for debug messages like network traffic, no data is present
				 so we'll have to take the command pointer */
		case IRCDEBUG_NET_IN:
			fprintf(stderr, "[IN] %s\n", command);
			break;
		case IRCDEBUG_NET_OUT:
			fprintf(stderr, "[OUT] %s", command);
			break;
		case IRCERR_UNHANDLED_SERVER_CMD:
			fprintf(stderr, "Unhandled server reply: '%d'\n", reply);
			break;
		case IRCERR_UNHANDLED_CLIENT_CMD:
			fprintf(stderr, "Unhandled client command: '%s'\n", command);
			break;
		case IRCERR_UNHANDLED_CTCP_CMD:
			fprintf(stderr, "Unhandled CTCP command: '%s'\n", command);
			break;
		default:
			fprintf(stderr, "[libsqirc:ERR] %lx: %s\n", irc_error, command);
			break;
		}
}

/* **************************************************************** */
/*                     client commands                              */
/* **************************************************************** */

SQIRC_FUNC_CMDCLIENT(internal_clientcmd_privmsg)
{
	struct client_command c_cmd;           /* client or ctcp command */
	char *message = data->message+1;
	
	bzero(&c_cmd, sizeof(c_cmd));
	
	/* handle ctcp commands */
	if (message[0] == 1 && message[strlen (message) - 1] == 1)
		{
			char cmd[IRC_BUFSIZE];
			int i;
			
			bzero(cmd, IRC_BUFSIZE);
			
			strncpy(cmd, message+1, strlen(message)-2);
			for (i=0;cmd[i] && cmd[i] != ' '; i++) {}
			cmd[i] = 0;
			
			sqirc_find_ctcp_command(cmd, 0, &c_cmd);
			
			if (!c_cmd.command[0])
				sqirc_error_handler(connection, IRCERR_UNHANDLED_CTCP_CMD,
														0, cmd, data);
			else
				c_cmd.callback(connection, command, data);
			
			return;
		}
	
	/* no ctcp command, now call developer's callback */
	sqirc_find_client_command(command, 1, &c_cmd);
	if (c_cmd.command[0] != 0)
		c_cmd.callback(connection, command, data);
}

SQIRC_FUNC_CMDCLIENT(internal_clientcmd_ping)
{
	struct client_command c_cmd;

	/*
		the developer provided function is only called,
		if the NO_AUTO_PONG flag is set; the default is to automatically
		reply to server PINGs
	*/
	if ( (connection->flags & SQIRC_NO_AUTO_PONG) == 0)
		sqirc_send(connection, "PONG :%s\r\n", data->message+1);
	else
		{
			sqirc_find_client_command(command, 1, &c_cmd);
			if (c_cmd.command[0] != 0)
				c_cmd.callback(connection, command, data);
		}
}

/* **************************************************************** */
/*                       ctcp commands                              */
/* **************************************************************** */

SQIRC_FUNC_CMDCLIENT(internal_ctcpcmd_ping)
{
	data->list[1][strlen(data->list[1])-1] = 0;
	sqirc_send(connection, "NOTICE %s :\001PING %s\001\r\n",
						 data->from.nick, data->list[1]);
}

/* **************************************************************** */
/*                     server commands                              */
/* **************************************************************** */

SQIRC_FUNC_CMDSERVER(internal_servercmd_motd)
{
	struct server_command s_cmd;

	if (connection->flags & SQIRC_SKIP_MOTD)
		return;

	sqirc_find_server_command(reply, 1, &s_cmd);
	if (s_cmd.reply != 0)
		s_cmd.callback(connection, reply, data);
	else if (connection->flags & SQIRC_DEBUG)
		sqirc_error_handler(connection, IRCERR_UNHANDLED_SERVER_CMD,
												reply, 0, data);
}
