#ifndef __HAVE_SQIRC_H
#define __HAVE_SQIRC_H

#ifdef BUFSIZE
#undef BUFSIZE                          /* force our size */
#endif
#define IRC_BUFSIZE 256
#define IRC_BUFNICK 16
#define IRC_BUFUSER 64

#include "replies.h"

struct irc_userdata {
	char username[IRC_BUFUSER];
	char nickname[IRC_BUFNICK];;
	char desc[IRC_BUFSIZE];
};

struct connection {
  int socket;
	int status;                           /* IRC_UNUSED,
																					 IRC_CONNECTED,
																					 IRC_UNCONNECTED */
	int input_id;                         /* GDK input id */
  time_t last_send;
  unsigned int bytes_sent;
	unsigned long flags;                  /* see below for possible values */
	struct irc_userdata userdata;         /* userdata for this connection */
	unsigned int pos;                     /* position in buffer */
	char buffer[2048];                    /* connection buffer */
  GSList *queue;
	void *data;                           /* ptr to application data  */
};

struct from_data {
	char token[IRC_BUFSIZE];                  /* will contain the whole token */
	char nick[IRC_BUFSIZE];                   /* nickname */
	char user[IRC_BUFSIZE];                   /* username */
	char host[IRC_BUFSIZE];                   /* hostname */
};

struct callback_data {
	char nick[IRC_BUFSIZE];               /* nickname */
	struct from_data from;                /* see above */
	char message[IRC_BUFSIZE];            /* everything after the last token */
	char **list;                          /* whitespace-splitted message words */
	int wcount;                           /* # of words in list */
	char buffer[2048];                    /* unparsed data */
};

struct server_command {
	int reply;
	void (*callback)(struct connection *, int, struct callback_data *);
};

struct client_command {
	char command[IRC_BUFSIZE];
	void (*callback)(struct connection *, char *, struct callback_data *);
};

/* some needed default values */
#define IRC_DEFAULT_PORT 6667
#define IRC_DEFAULT_USER "unknown"
#define IRC_DEFAULT_NICK "unknown"
#define IRC_DEFAULT_DESC "unknown"

#define IRC_SENDMAX       256       /* max # of bytes to send out at once */
#define IRC_QUEUETIMEOUT 2500       /* msec's to wait before sending again */

/* stati of various data fields */
#define IRC_UNUSED         -1                 /* socket status */
#define IRC_UNCONNECTED    0                  /* socket status */
#define IRC_CONNECTED      1                  /* socket status */ 

/* callback function definitions */
#define SQIRC_FUNC_CMDSERVER(x) void x(struct connection *connection, int reply, struct callback_data *data)
#define SQIRC_FUNC_CMDCLIENT(x) void x(struct connection *connection, char *command, struct callback_data *data)
/* error handler design */
#define SQIRC_FUNC_ERRORHANDLER(x) void x(struct connection *connection, ulong irc_error, int reply, char *command, struct callback_data *data)

/* convenience definitions for command lists */
#define LAST_COMMAND 	{ "", NULL }
#define CMDIGNORE(x) { x, NULL }

/* convenience definitions for GSList handling */
#define IRC_ADD_LIST_ITEM(list, item) list = g_slist_append(list, item)
#define IRC_DEL_LIST_ITEM(list, item) list = g_slist_remove(list, item)

/* errors returned by sqirc_connect */
#define IRCERR_BASE                     0xF000
#define IRCERR_INVALID_HOSTNAME        (0x0001 + IRCERR_BASE)
#define IRCERR_LOOKUP_FAIL             (0x0002 + IRCERR_BASE)
#define IRCERR_NOSOCKET                (0x0003 + IRCERR_BASE)
#define IRCERR_NOCONNECTION            (0x0004 + IRCERR_BASE)
#define IRCERR_NOTCONNECTED            (0x0005 + IRCERR_BASE)
#define IRCERR_BUFFEROVERFLOW          (0x0006 + IRCERR_BASE)
#define IRCERR_DATAMISSING             (0x0050 + IRCERR_BASE)
/* errors handled by your error handler, if any */
#define IRCERR_UNHANDLED_CTCP_CMD      (0x0101 + IRCERR_BASE)
#define IRCERR_UNHANDLED_CLIENT_CMD    (0x0102 + IRCERR_BASE)
#define IRCERR_UNHANDLED_SERVER_CMD    (0x0103 + IRCERR_BASE)
/* non-error messages, yet handled by the error handler */
#define IRCDEBUG_NET_IN                  0x0001
#define IRCDEBUG_NET_OUT                 0x0002

/* flags for struct connection.flags */
#define SQIRC_DEBUG               1<<0    /* print network traffic to stderr */
#define SQIRC_SKIP_MOTD           1<<1    /* skip message of the day */
#define SQIRC_NO_AUTO_PONG        1<<2    /* dont automatically answer pings */

/* callback types for sqirc_register_callback */
#define IRC_FUNC_CLEANUP         1         /* cleanup function */
#define IRC_FUNC_READDATA        2         /* non-default socket read-data */
#define IRC_FUNC_ERRORHANDLER    3         /* error handler */

/* and finally the functions we report the outside */
void sqirc_initialize(struct server_command *,
											struct client_command *,
											struct client_command *);

void sqirc_register_client_commands(struct client_command *);
void sqirc_register_ctcp_commands(struct client_command *);
void sqirc_register_server_commands(struct server_command *);
void sqirc_register_callback(int, void (*callback)());

int sqirc_connect(const char *,               /* hostname */
									const int,                  /* port */
									const int,                  /* flags */
									struct irc_userdata *,
									struct connection **);
void sqirc_close_connection(struct connection *connection);
void sqirc_close_all_connections();
int sqirc_process_buffer(struct connection *);
int sqirc_send(struct connection *, char *, ...);


#endif /* __HAVE_SQIRC_H */
