/* Squaroid
 * Copyright (C) 1999,2000 Lukas Schrder
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "squaroid.h"

#if DEBUG || NET_DEBUG
static GtkWidget *debug_text;
#endif

void chatroom_localcommand(Chatroom *, char *command);
void chatroom_say(GtkWidget *, gpointer);
void cb_select_row(GtkWidget *, gint, gint, GdkEventButton *, gpointer);
gint playerlist_sort(GtkWidget *, gconstpointer, gconstpointer);

/* local command callbacks; called from chatroom_localcommand */
#define LC_FUNC(_x) void _x(Chatroom *chatroom, char *rawmsg, char **list, int count)
static LC_FUNC(lc_msg);
static LC_FUNC(lc_quit);
static LC_FUNC(lc_part);
static LC_FUNC(lc_score);
static LC_FUNC(lc_allscores);
static LC_FUNC(lc_solo);
static LC_FUNC(lc_disconnect);
static LC_FUNC(lc_join);
static LC_FUNC(lc_save);
static LC_FUNC(lc_restore);
static LC_FUNC(lc_me);
static struct local_command {
	char *name;
	void (*callback)(Chatroom *, char *, char **, int);
} local_commands[] =
{ { "/msg",           lc_msg },
	{ "/notice",        lc_msg },
	{ "/score",         lc_score },
	{ "/allscores",     lc_allscores },
	{ "/quit",          lc_quit },
	{ "/part",          lc_part },
	{ "/exit",          lc_quit },
	{ "/solo",          lc_solo },
	{ "/disconnect",    lc_disconnect },
	{ "/join",          lc_join },
	{ "/save",          lc_save },
	{ "/restore",       lc_restore },
	{ "/me",            lc_me },
	{ "",               0 }
};

/* this will create and show our toplevel window */
Chatroom *chatroom_create(struct connection *connection) 
{
	Chatroom *chatroom = malloc(sizeof(Chatroom));
	GtkWidget *win;
	
	bzero(chatroom, sizeof(Chatroom));
	
	if (find_chatroom_by_connection(0) == 0)     /* no dummy yet, create one */
		return chatroom;

	/* dont associate chatroom to connection if it's a sub-room */
	if (connection != NULL && connection->data == NULL)
		connection->data = chatroom;
	chatroom->connection = connection;
	
  chatroom->table = gtk_table_new(10, 10, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(chatroom->table), 2);
  gtk_table_set_col_spacings(GTK_TABLE(chatroom->table), 5);
	
  chatroom->text = gtk_text_new(NULL, NULL);
  gtk_widget_set_usize(chatroom->text, 390, 350);
  chatroom->text_style = sq_get_widget_style();
  gtk_widget_set_style(chatroom->text, chatroom->text_style);
	
	win = gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win), 
																 GTK_POLICY_NEVER,
																 GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(win), chatroom->text);
  gtk_table_attach_defaults(GTK_TABLE(chatroom->table), win, 0, 6, 0, 6);
	
	chatroom->playerlist = gtk_clist_new(4);
	gtk_clist_set_column_width(GTK_CLIST(chatroom->playerlist), 0, 20);
  gtk_clist_set_column_width(GTK_CLIST(chatroom->playerlist), 1, 20);
  gtk_clist_set_column_width(GTK_CLIST(chatroom->playerlist), 2, 90);
  gtk_clist_set_auto_sort(GTK_CLIST(chatroom->playerlist), TRUE);
  gtk_clist_set_compare_func(GTK_CLIST(chatroom->playerlist),
														 (GtkCListCompareFunc)playerlist_sort);
	win = gtk_scrolled_window_new(0, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win), 
																 GTK_POLICY_NEVER,
																 GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(win), chatroom->playerlist);
  gtk_table_attach_defaults(GTK_TABLE(chatroom->table),	win, 6, 8, 0, 6);
	
  chatroom->chatline = gtk_entry_new_with_max_length(128);
  gtk_table_attach(GTK_TABLE(chatroom->table), chatroom->chatline,
									 0, 7, 6, 7, GTK_FILL | GTK_EXPAND, 0, 0, 0);
	
  chatroom->cmdsay = gtk_button_new_with_label("Say!");
  gtk_table_attach(GTK_TABLE(chatroom->table), chatroom->cmdsay,
									 7, 8, 6, 7, GTK_FILL | GTK_EXPAND, 0, 0, 0);

  gtk_signal_connect(GTK_OBJECT(chatroom->cmdsay), "clicked",
										 GTK_SIGNAL_FUNC(chatroom_say), chatroom);
  gtk_signal_connect(GTK_OBJECT(chatroom->chatline), "activate",
										 GTK_SIGNAL_FUNC(chatroom_say), chatroom);
  gtk_signal_connect(GTK_OBJECT(chatroom->playerlist), "select_row",
										 GTK_SIGNAL_FUNC(cb_select_row), chatroom);
	
  chatroom->tablabel = gtk_label_new("<none>");
	gtk_widget_set_style(chatroom->tablabel, greenstyle);
  gtk_widget_show(chatroom->tablabel);
  sq_notebook_append_page(chatroom->table, chatroom->tablabel);
	
  gtk_widget_show_all(chatroom->table);
	return chatroom;
}

void chatroom_destroy(Chatroom *chatroom)
{
	Chatroom *subroom = 0;
	
	chatroom->channel[0] = 0;
	if (chatroom->type == CHATROOM_SUBROOM)
		{
			chatroom_player_clearlist(chatroom);
			DEL_LIST_ITEM(CHATROOM(chatroom->connection)->rooms, chatroom);
		}
	else if (chatroom->rooms != 0)
		{
			/* create a new master from the first subroom listed */
			subroom = (Chatroom *)chatroom->rooms->data;
			subroom->type = CHATROOM_MAIN;
			chatroom->connection->data = subroom;
			
			DEL_LIST_ITEM(chatroom->rooms, subroom);
			ADD_LIST_ITEM(chatroom_list, subroom);
			
			subroom->rooms = chatroom->rooms;
			subroom->players = chatroom->players;
			subroom->games = chatroom->games;
		}
	else if (chatroom->rooms == 0)
		{
			/* last chatroom of this connection */
			sqirc_close_connection(chatroom->connection);
			
			chatroom->connection = 0;
		}
	
	chatroom_player_clearlist(chatroom);
	if ( (chatroom->type != CHATROOM_SUBROOM)
			 && (g_slist_length(chatroom_list) <= 2) )
		return;                   /* dont remove tab if it's the last window */
	
	/* take away any visual associated with this chatroom */
	gtk_widget_destroy(chatroom->table);
	
	if (current_tablabel == chatroom->tablabel)
		gtk_notebook_set_page(GTK_NOTEBOOK(top_notebook), 0);
	
	if (g_slist_index(chatroom_list, chatroom) != -1)
		DEL_LIST_ITEM(chatroom_list, chatroom);
	
	free(chatroom);                          /* free all malloc's */
}

void chatroom_join(Chatroom *chatroom, const char *channel)
{
	gtk_label_set_text(GTK_LABEL(chatroom->tablabel), channel);
	
	bzero(chatroom->channel, SMALLBUF);
	strncpy(chatroom->channel, channel, SMALLBUF-1);
}

void chatroom_create_debugwindow()
{
#if DEBUG || NET_DEBUG
	GtkWidget *vbox;
	GtkStyle *text_style;
	
	vbox = gtk_vbox_new(0,0);
	
  debug_text = gtk_text_new(NULL, NULL);
	gtk_box_pack_start(GTK_BOX(vbox), debug_text, TRUE, TRUE, 0);
  gtk_widget_set_usize(debug_text, 390, 350);
  text_style = sq_get_widget_style();
  gtk_widget_set_style(debug_text, text_style);
	
  chatroom_debuglabel = gtk_label_new("Debug");
	gtk_widget_set_style(chatroom_debuglabel, blackstyle);
  gtk_widget_show(chatroom_debuglabel);
  sq_notebook_append_page(vbox, chatroom_debuglabel);
	
	gtk_widget_show_all(vbox);
#endif
}

void chatroom_chatmessage(Chatroom *chatroom, char *fmt, ...)
{
	char data[BUFSIZE*2];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(data, BUFSIZE*2, fmt, ap);
	va_end(ap);
	
  PrintText( chatroom->text, data);
	
	if (current_tablabel != chatroom->tablabel)
    gtk_widget_set_style(chatroom->tablabel, greenstyle);
}

void chatroom_debugmessage(char *fmt, ...)
{
#if DEBUG || NET_DEBUG
	char data[BUFSIZE*2];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(data, BUFSIZE*2, fmt, ap);
	va_end(ap);
	
  PrintTextDebug(data);
#endif
}

void chatroom_player_add(Chatroom *chatroom, const char *nick, guint score)
{
	Player *player = 0;
  GdkImlibImage *im = 0;
  int row;
  char buf[64];
  gchar *content[4];
  char buffer[512];
	
  content[0] = '\0'; content[1] = '\0';
  content[2] = (char *)nick;
  sprintf(buf, "%d", score);
  content[3] =  buf;

	/* skip if it's already listed */
  if (chatroom_player_inlist(chatroom, nick))
    return;
	
  if ( (player = find_player_by_nick(chatroom->connection, nick)) == 0 )
    {
      player = sq_player_create(chatroom);
      strncpy(player->nickname, nick, 63);
    }
	else
		snprintf(buf, 64, "%d", player->score);
  row = gtk_clist_append(GTK_CLIST(chatroom->playerlist), content);
	
  /* check if ok to paint the invite & observe pixmaps
   * and assume that the player is using a valid squaroid client
   * we'll check whether this assumptions is right as soon as
   * a request of any kind is sent to this player (via CTCP version)
   */
  if (strcmp(player->nickname, preferences.nickname)
      && strcasecmp(player->nickname, "squaroid"))
    {
      GFXPATH(preferences.theme_data.invite, buffer);
      im = gdk_imlib_load_image(buffer);
      gdk_imlib_render(im, im->rgb_width, im->rgb_height);
      gtk_clist_set_pixmap(GTK_CLIST(chatroom->playerlist), row, 0,
													 gdk_imlib_move_image(im),
													 gdk_imlib_move_mask(im));
			
      GFXPATH(preferences.theme_data.observe,buffer);
      im = gdk_imlib_load_image(buffer);
      gdk_imlib_render(im, im->rgb_width, im->rgb_height);
      gtk_clist_set_pixmap(GTK_CLIST(chatroom->playerlist), row, 1,
													 gdk_imlib_move_image(im),
													 gdk_imlib_move_mask(im));
    }
}

gint chatroom_player_inlist(Chatroom *room, const char *name)
{
	Chatroom *chatroom = room;
  char *player = 0;
  int row, len;
	
  for (row=0; row<GTK_CLIST(chatroom->playerlist)->rows; row++)
		{
			gtk_clist_get_text(GTK_CLIST(chatroom->playerlist), row, 2, &player);
			len = MAX(strlen(name), strlen(player));
			if (!strncmp(name, player, len))
				/* to be true if row == 0;
				 * if this is to be used, decrease by 1 again!!!! */
				return row+1; 
		}
	
  return 0;
}

void chatroom_player_remove(Chatroom *chatroom, char *name, char *reason)
{
  int row;
  char *player;
	GSList *list;
	
	/* handle the main room first */
  for (row=0;row<GTK_CLIST(chatroom->playerlist)->rows;row++)
		{
			gtk_clist_get_text(GTK_CLIST(chatroom->playerlist), row, 2, &player);
			if (!strcmp(name, player))
				{
					if (reason)
						chatroom_chatmessage(chatroom,
																 STARTOFF " Player \0033%s\003 quit "
																 "(\0032%s\003)\n", name, reason);
					gtk_clist_remove(GTK_CLIST(chatroom->playerlist), row);
				}
		}
	
	/* now walk through the list of subrooms */
	list = chatroom->rooms;
	while (list)
		{
			Chatroom *room = (Chatroom *)list->data;
			
			for (row=0;row<GTK_CLIST(room->playerlist)->rows;row++)
				{
					gtk_clist_get_text(GTK_CLIST(room->playerlist), row, 2, &player);
					if (!strcmp(name, player))
						{
							if (reason)
								chatroom_chatmessage(room,
																		 STARTOFF " Player \0033%s\003 quit "
																		 "(\0032%s\003)\n", name, reason);
							gtk_clist_remove(GTK_CLIST(room->playerlist), row);
						}
				}
			
			list = list->next;
		}
}

void chatroom_player_update(Chatroom *chatroom, Player *player)
{
  int row;
  char *content;
  char buf[64];
  
  sprintf(buf, "%d", player->score);
  for(row=0;row<GTK_CLIST(chatroom->playerlist)->rows; row++)
		{
			gtk_clist_get_text(GTK_CLIST(chatroom->playerlist), row, 2, &content);
			if (!strcmp(player->nickname, content))
				gtk_clist_set_text(GTK_CLIST(chatroom->playerlist), row, 3, buf);
		}
	
  gtk_clist_sort(GTK_CLIST(chatroom->playerlist));
}

void chatroom_player_mark_invalid(Chatroom *chatroom, int row)
{
	if (row < 0)
		return;
  gtk_clist_clear_pixmap(GTK_CLIST(chatroom->playerlist), row, 0);
  gtk_clist_clear_pixmap(GTK_CLIST(chatroom->playerlist), row, 1);
}

LC_FUNC(lc_msg)
{
	char message[BUFSIZE];
	int i;

	bzero(message, BUFSIZE);
	
	if (!list[0] || !list[1])
		{
			chatroom_chatmessage(chatroom, 
													 STARTON " Usage: /MSG <nick> <message>\n");
			return;
		}
	
	for(i = 1; list[i] && (strlen(message)+strlen(list[i])) < BUFSIZE; i++)
		{
			strcat(message, list[i]);
			strcat(message, " ");
		}
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		{
			sq_print_on_current_chatwindow(STARTON " You are not connected\n");
			return;
		}
	
	net_send_privmsg(chatroom->connection, list[0], message);
	
}

LC_FUNC(lc_quit)
{
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		sq_print_on_current_chatwindow(STARTON " You are not connected\n");
	else 
		sqirc_close_connection(chatroom->connection);
	show_dialog_connect();
}

LC_FUNC(lc_part)
{
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		{
			sq_print_on_current_chatwindow(STARTON " You are not connected\n");
			return;
		}

	sqirc_send(chatroom->connection, "PART %s\r\n", chatroom->channel);
	chatroom_destroy(chatroom);
}

LC_FUNC(lc_score)
{
	if (!list[0])
		{
			chatroom_chatmessage(chatroom, STARTON " Usage: /SCORE <nick>\n"); 
			return;
		}
	if (!chatroom_player_inlist(chatroom, "Squaroid"))
		{
			chatroom_chatmessage(chatroom,
													 STARTON " 'Squaroid'-Bot is not present.\n");
			return;
		}	
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		{
			sq_print_on_current_chatwindow(STARTON " You are not connected\n");
			return;
		}

	sqirc_send(chatroom->connection, "NOTICE Squaroid :SCORE %s\r\n", list[1]);
}

LC_FUNC(lc_allscores)
{
	if (!chatroom_player_inlist(chatroom, "Squaroid"))
		{
			chatroom_chatmessage(chatroom,
													 STARTON " 'Squaroid'-Bot is not present.\n");
			return;
		}
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		{
			sq_print_on_current_chatwindow(STARTON " You are not connected\n");
			return;
		}
	
	sqirc_send(chatroom->connection, "NOTICE Squaroid :ALLSCORES\r\n");
}

LC_FUNC(lc_solo)
{
	menucallback_start_sologame(NULL, NULL);
}

LC_FUNC(lc_disconnect)
{
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		{
			sq_print_on_current_chatwindow(STARTON " You are not connected\n");
			return;
		}

	sqirc_close_connection(chatroom->connection);
}

LC_FUNC(lc_join)
{
	Chatroom *newroom;
	if (!list[0])
		{
			sq_print_on_current_chatwindow(STARTON " Usage: /JOIN <channel>\n"); 
			return;
		}
	if (list[0][0] != '#')
		{
			sq_print_on_current_chatwindow(STARTON " <channel> must start with #\n");
			return;
		}
	if (chatroom->connection == 0
			|| chatroom->connection->status != IRC_CONNECTED)
		{
			sq_print_on_current_chatwindow(STARTON " You are not connected\n");
			return;
		}
	if ( (find_chatroom_by_channel(chatroom->connection, list[0])) )
		{
			sq_print_on_current_chatwindow(STARTON " You have already joined this "
																		 "room\n");
			return;
		}
	
	newroom = chatroom_create(chatroom->connection);
	newroom->connection = chatroom->connection;
	
	ADD_LIST_ITEM(chatroom->rooms, newroom);
	sqirc_send(chatroom->connection, "JOIN :%s\r\n", list[0]);
	
	/* it's a subroom, so just assign the master's lists too */
	newroom->games = chatroom->games;
	newroom->players = chatroom->players;
	newroom->type = CHATROOM_SUBROOM;
	
	chatroom_join(newroom, list[0]);
}

LC_FUNC(lc_save)
{
	if (chatroom->gamehost[0] != 0)
		{
			sqirc_send(chatroom->connection, "PRIVMSG %s :\001SAVE\001\r\n",
								 chatroom->gamehost);
			chatroom_chatmessage(chatroom,
													 STARTON " save request has been sent out\n");
		}
	else
		{
			chatroom_chatmessage(chatroom,
													 STARTON " save request cannot be sent out. No Host-Bot found.\n");
		}
}

LC_FUNC(lc_restore)
{
	if (chatroom->gamehost[0] != 0 && preferences.playerid[0] != 0)
		{
			sqirc_send(chatroom->connection, "PRIVMSG %s :\001RESTORE %s -\001\r\n",
								 chatroom->gamehost, preferences.playerid);
			chatroom_chatmessage(chatroom,
													 STARTON " restore request has been sent out\n");
		}
	else if (chatroom->gamehost[0] == 0)
		chatroom_chatmessage(chatroom,
												 STARTON " restore request cannot be sent out. No Host-Bot found.\n");
	else if (preferences.playerid[0] == 0)
		chatroom_chatmessage(chatroom,
												 STARTON " restore request cannot be sent out. "
												 "You dont have a player-id yet.\n   "
												 "Use /save to save your score and request a "
												 "player-id\n");
	
}

LC_FUNC(lc_me)
{
	if (chatroom->connection->status == IRC_CONNECTED)
		{
			sqirc_send(chatroom->connection, "PRIVMSG %s :\001ACTION %s\001\r\n",
								 chatroom->channel, rawmsg);
			chatroom_chatmessage(chatroom, ACTIONMESSAGE,
													 preferences.nickname, rawmsg);
		}
}

void chatroom_localcommand(Chatroom *chatroom, char *command)
{
	char name[BUFSIZE];
	char words[16][BUFSIZE], **list;
	char rawmsg[BUFSIZE];
	char *cp1, *cp2, *eos;             /* generic char pointer */
	int wcount;                        /* # of words found */
	int i;
	struct local_command lcmd;
  
	bzero(name,BUFSIZE);
	bzero(words, sizeof(words));

	cp1 = command;
	cp2 = name;
	eos = cp2 + BUFSIZE + 1;           /* end of string */

#define SKIPWHITESPACE(_cp) while (isspace(*_cp) && *_cp != 0) _cp++
	
	while ( (*cp1 != 0) && !isspace(*cp1) && (cp2 != eos) )
		*cp2++ = *cp1++;
	*cp2 = 0;
	
	SKIPWHITESPACE(cp1);
	
	/* now check if we can find that command */
	bzero(&lcmd, sizeof(lcmd));
	for (i=0; local_commands[i].name[0]; i++)
		{
			if ( !strcmp(local_commands[i].name, name) )
				{
					lcmd = local_commands[i];
					break;
				}
		}
	SKIPWHITESPACE(cp1);
	
	if (!lcmd.name)
		{
			chatroom_chatmessage(chatroom,
													 STARTON " Command '%s' unknown\n", name);
			return;
		}

	strncpy(rawmsg, cp1, BUFSIZE);
	wcount = 0;
	while (*cp1 != 0)
		{
			cp2 = words[wcount++];
			eos = cp2 + BUFSIZE + 1;
			while ( cp2 != eos && *cp1 != 0 && !isspace(*cp1) )
				*cp2++ = *cp1++;
			*cp2 = 0;
			SKIPWHITESPACE(cp1);
		}
	
	list = malloc(sizeof(char*) * (wcount+1));
	for (i=0; i<wcount; i++)
		list[i] = strdup(words[i]);
	list[wcount] = 0;
	
	lcmd.callback(chatroom, rawmsg, list, wcount);
	
	for (i=0; i<wcount; i++)
		free(list[i]);
	free(list);
}

void chatroom_player_clearlist(Chatroom *chatroom)
{
  gtk_clist_clear(GTK_CLIST(chatroom->playerlist));
}

void chatroom_freeze(Chatroom *chatroom)
{
  gtk_clist_freeze(GTK_CLIST(chatroom->playerlist));
}

void chatroom_thaw(Chatroom *chatroom)
{
  gtk_clist_thaw(GTK_CLIST(chatroom->playerlist));
}

/* CALLBACKS */
void chatroom_say(GtkWidget *w, gpointer data)
{
  char *line;
	Chatroom *chatroom = (Chatroom *)data;
  struct connection *connection = chatroom->connection;
	
  line = gtk_entry_get_text(GTK_ENTRY(chatroom->chatline));
  if (line[0] == '/')
		{ 
			chatroom_localcommand(chatroom, line);
			gtk_entry_set_text(GTK_ENTRY(chatroom->chatline), "");
			return;
		}
  if (line[0] == '\r' || line[0] == '\n' || line[0] == '\0')
    return;
	if ((sqirc_send(connection,"PRIVMSG %s :%s\r\n",chatroom->channel,line)) > 0)
		chatroom_chatmessage(chatroom,
												 CHATMESSAGE, preferences.nickname,
												 gtk_entry_get_text(GTK_ENTRY(chatroom->chatline)));
	gtk_entry_set_text(GTK_ENTRY(chatroom->chatline), "");
}

void cb_select_row(GtkWidget *w, gint row, gint col,
									 GdkEventButton *e, gpointer data)
{
  Game *game = 0;
  Player *e_player = 0;
  char *player = 0;
  time_t now = time(0);
	Chatroom *chatroom = (Chatroom *)data;
  struct connection *connection = chatroom->connection;
	
  gtk_clist_unselect_row(GTK_CLIST(w), row, col);
  if (col > 1) return;
  gtk_clist_get_text(GTK_CLIST(w), row, 2, &player);
  
  /* we don't do anything if our entry has been clicked */
  if (!strcmp(player, preferences.nickname) || !strcasecmp(player, "squaroid"))
    return;
	
  e_player = find_player_by_nick(connection, player);
	assert(e_player != 0);
  if (e_player->valid && e_player->last_update
			&& e_player->last_update > now-10)
		{
			fprintf(stderr, "e_player[%s] lastupdate > now-10\n", player);
			chatroom_chatmessage(CHATROOM(connection), 
													 STARTOFF " You may only invite a player once "
													 "within 10 seconds.\n");
			return;
		}
  
  if (!(game = find_game_by_opponent(connection, player)))
		{
			game = game_create(connection);
			strcpy(game->opponent, player);
			game->e_opponent = e_player;
		}
	else
		{
			chatroom_chatmessage(CHATROOM(connection), 
													 STARTOFF " You are already playing against "
													 "%s\n", player);
			return;
		}
			
	fprintf(stderr, "select_row: Game %p\n", game);
  if (game->pending_invite || game->pending_observe)
		{
			chatroom_chatmessage(CHATROOM(connection), 
													 STARTOFF " There is another invitation still "
													 "pending!\n");
			return;
		}
	
  switch(col)
		{
		case 0:
			game->pending_invite = 1;
			
			if (sq_player_clientinfo(game->e_opponent) == TRUE)
				game_play_request(game);
			
			break;
		case 1:
			game->name[0] = 0;
			game->pending_observe = 1;
			net_send_ctcp_version(connection, player);
			timeout_add_pending_observe_request(game);
			break;
		default:
			break;
		};
}

gint playerlist_sort(GtkWidget *w, gconstpointer ptr1, gconstpointer ptr2)
{
  GtkCListRow *row1 = (GtkCListRow *)ptr1;
  GtkCListRow *row2 = (GtkCListRow *)ptr2;
  char *text1 = GTK_CELL_TEXT(row1->cell[3])->text;
  char *text2 = GTK_CELL_TEXT(row2->cell[3])->text;
  int a, b;

  a = atoi(text1);
  b = atoi(text2);

  if (a < b) return 1;
  else if (a > b) return -1;
  return 0;
}

void PrintTextRaw (GtkWidget * textwidget,
									 unsigned char *text,
									 GdkFont * fontnorm,
									 GdkFont * fontbold)
{
	int esc = FALSE;
	int comma = FALSE;
	unsigned char buf[4096];
	unsigned char num[4];
	int bcol = 1, col = 0, j = 0, i = 0, k = 0;
	int scroll = FALSE;
	GtkAdjustment *adj;
	GdkFont *font = fontnorm;
	int bold = FALSE;
	
	bzero(buf, 4096);
	adj = (GTK_TEXT (textwidget))->vadj;
	if (adj->value == adj->upper - adj->lower - adj->page_size)
		scroll = TRUE;
	
	gtk_text_freeze (GTK_TEXT (textwidget));
	while (1)
		{
      if (esc)
				{
					if (!isdigit (text[i]))
						{
							bcol = 1;
							col = 0;
							goto norm;
						} else
							{
								while (text[i] >= '0' && text[i] <= '9' && k < 2)
									{
										num[k] = text[i];
										k++;
										i++;
									}
								num[k] = 0;
								k = 0;
								switch (text[i])
									{
									case ',':
										comma = TRUE;
										col = atoi (num);
										if (col < 0 || col > 15)
											col = 1;
										break;
									default:
										if (comma)
											{
												comma = FALSE;
												bcol = atoi (num);
												if (bcol < 0 || bcol > 15)
													bcol = 1;
											} else
												{
													col = atoi (num);
													if (col < 0 || col > 15)
														col = 0;
												}
										if (bcol == 1 && col == 1)
											col = 0;
										if (bcol == 0 && col == 0)
											col = 1;
										goto norm;
									}
							}
				} else
					{
					norm:
						esc = FALSE;
						switch (text[i])
							{
							case 0:
								goto jump;
							case 3:               /* CTRL-C */
								buf[j] = 0;
								if (j > 0)
									{
										gtk_text_insert (GTK_TEXT (textwidget), font,
																		 &colors[col], &colors[bcol], buf, j);
										j = 0;
									}
								esc = TRUE;
								k = 0;
								break;
							case '\007':          /* beep */
								break;
							case '\026':          /* REVERSE */
							case '\037':          /* UNDERLINE */
							case '\002':          /* BOLD */
								buf[j] = 0;
								if (j > 0)
									{
										gtk_text_insert (GTK_TEXT (textwidget), font,
																		 &colors[col], &colors[bcol], buf, j);
										j = 0;
									}
								if (bold)
									{
										bold = FALSE;
										font = fontnorm;
									} else
										{
											bold = TRUE;
											font = fontbold;
										}
								break;
							case '\017':          /* ALL OFF */
								buf[j] = 0;
								if (j > 0)
									{
										gtk_text_insert (GTK_TEXT (textwidget), font,
																		 &colors[col], &colors[bcol], buf, j);
										bcol = 1;
										col = 0;
										bold = FALSE;
										j = 0;
									}
								font = fontnorm;
								break;
							default:
								esc = FALSE;
								comma = FALSE;
								buf[j] = text[i];
								j++;
								if (j == 4095)
									j = 4094;
							}
					}
      i++;
			if (text[i] == '\r')
				i++;
		}
 jump:
	if (j)
		{
      gtk_text_insert (GTK_TEXT (textwidget), font,
											 &colors[col], &colors[bcol], buf, j);
      gtk_text_thaw (GTK_TEXT (textwidget));
      if (scroll)
				gtk_adjustment_set_value (adj,
																	adj->upper - adj->lower - adj->page_size);
		} else
			{
				gtk_text_thaw (GTK_TEXT (textwidget));
			}
}

void PrintText (GtkWidget *textwidget, char *message)
{
	PrintTextRaw(textwidget, message,
							 gdk_font_load("fixed"), gdk_font_load("fixed"));
}

void PrintTextDebug (char *message)
{
#if DEBUG || NET_DEBUG
	PrintTextRaw(debug_text, message,
							 gdk_font_load("fixed"), gdk_font_load("fixed"));
	gtk_widget_set_style(chatroom_debuglabel, greenstyle);
#endif
}
