/* Squaroid
 * Copyright (C) 1999 Lukas Schrder
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "squaroid.h"

#include <gdk/gdkprivate.h>

/* game_ai.c */
extern int ai_move();
extern void board_init();
extern void score_init();
extern int ai_is_playing;

	/* game is over if anybody got 150 points and it 15+ in the lead */
#define IS_GAME_OVER 	game->game_over = (game->game_over \
                          || ((game->p1_score >= 150 \
                               && game->p1_score-game->p2_score >= 15)\
			       || (game->p2_score >= 150\
			       && game->p2_score-game->p1_score >= 15)))

void *game_alloc(Game *game, size_t size)
{
	void *mem = malloc(size);
	ADD_LIST_ITEM(game->memory, mem);
	
	bzero(mem, size);
	return mem;
}

void game_free(Game *game, void *ptr)
{
	DEL_LIST_ITEM(game->memory, ptr);
	free(ptr);
}
																				 
/* game_create() sets up a clean Game for a new session
 * to be freed using game_cleanup(Game *)
 */
Game *game_create(struct connection *connection)
{
  Game *game = malloc(sizeof(Game));
	
  memset(game, 0, sizeof(Game));
	if (connection && connection != CHATROOM_DUMMY)
		ADD_LIST_ITEM(CHATROOM(connection)->games, game);
	else if (connection == CHATROOM_DUMMY)
		ADD_LIST_ITEM((find_chatroom_by_connection(CHATROOM_DUMMY))->games, game);
	
  strcpy(game->name, preferences.nickname);
  game->options.first_player = PLAYER1;
  game->options.move_limit = 0;
  game->options.score_limit = 150;
	game->options.timeout = LIMIT_OFF;
	
  game->timeout_play = 0;
  game->timeout_observe = 0;
	game->old_lines = 0;
	game->old_stones = 0;
	game->dialog = 0;
	game->invited = NOT_INVITED;
  game->e_me = find_player_by_nick(connection, preferences.nickname);
	game->connection = connection;
	
	return game;
}

void game_cleanup(Game *game)
{
	
	timeout_del_movelimit(game);
  if (current_tablabel == game->tablabel)
    current_tablabel = 0;

#define SAFE_DESTROY(x) if (x) gtk_widget_destroy(x)

	SAFE_DESTROY(game->small_red);
	SAFE_DESTROY(game->small_blue);
	SAFE_DESTROY(game->spot_red);
	SAFE_DESTROY(game->spot_red_last);
	SAFE_DESTROY(game->spot_blue);
	SAFE_DESTROY(game->spot_blue_last);
	SAFE_DESTROY(game->bg_marker);
	SAFE_DESTROY(game->bg_tile);
	SAFE_DESTROY(game->p1_name);
	SAFE_DESTROY(game->p2_name);
	SAFE_DESTROY(game->p1_score_l);
	SAFE_DESTROY(game->p2_score_l);
  
  if (game->dialog_shown && GTK_IS_WIDGET(game->dialog))
		{
			gtk_widget_destroy(game->dialog);
			game->dialog_shown = FALSE;
			game->dialog = 0;
		
			while (gtk_events_pending())
				gtk_main_iteration();
		}
	
  if (GTK_IS_WIDGET(game->vbox))
		{
			gtk_widget_destroy(game->vbox);
			game->vbox = 0;
		
			while (gtk_events_pending())
				gtk_main_iteration();
		}
	
	sq_cleanup_simple_glist(game->old_lines);
  sq_cleanup_simple_glist(game->old_stones);
	
  game->pending_invite = 0;
  game->pending_observe = 0;
	
  if (game->timeout_play)
		timeout_del_pending_play_request(game);
	
  if (game->timeout_observe)
		timeout_del_pending_observe_request(game);
	
  if (game->connection && CHATROOM(game->connection)->games)
		DEL_LIST_ITEM(CHATROOM(game->connection)->games, game);
	
	/* cleanup memory alloc's */
	while (game->memory)
		game_free(game, game->memory->data);
	
	game_play_request_abort(game);
	
	memset(game, 0, sizeof(*game));
  free(game);
  game = 0;
	
  while (gtk_events_pending())
    gtk_main_iteration();
  
}

void game_cleanup_all(struct connection *connection)
{
	Game *game;
  while (CHATROOM(connection)->games)
		{
			game = (Game *)(CHATROOM(connection)->games)->data;
			game_cleanup(game);
		}
}

/* cleanup the Game */
void
game_reset_values (Game * game)
{
  gint x, y;
  char _p1_n[MAX_PLAYER_NAME + 1];
  char _p2_n[MAX_PLAYER_NAME + 1];
  enum _player me = PLAYER1, first = PLAYER1;

  if (game->mode == GAME_OBSERVE)
    return;

  if (game->invited == BEEN_INVITED)
    me = PLAYER2;
  else if (game->invited == HAVE_INVITED)
    me = PLAYER1;

  if (game->options.first_player == YOU_FIRST)
    first = PLAYER1;
  else
    first = PLAYER2;

  game->current_player = first;
  game->current_move = 0;
  game->p1_score = 0;
  game->p2_score = 0;
  game->who_am_i = me;
  game->playing = 1;

  for (x = 0; x < 8; x++)
    for (y = 0; y < 8; y++)
      game->points[x][y] = NONE;
  for (x = 0; x < 65; x++)
    game->moves[x].owner = NONE;

  sq_cleanup_simple_glist (game->old_lines);
  sq_cleanup_simple_glist (game->old_stones);
  game->old_lines = 0;
  game->old_stones = 0;

  if (game->mode == GAME_AI)
    {
      board_init ();
      score_init ();
    }

  /* reset names & styles */
  game_get_names (game, _p1_n, _p2_n);

  if (game->p1_name)
    {
      if (game->current_player == PLAYER1)
	gtk_widget_set_style (game->p1_name, greenstyle);
      else
	gtk_widget_set_style (game->p1_name, blackstyle);
    }
  if (game->p2_name)
    {
      if (game->current_player == PLAYER2)
	gtk_widget_set_style (game->p2_name, greenstyle);
      else
	gtk_widget_set_style (game->p2_name, blackstyle);
    }

  /* reset scores */
  if (game->p1_score_l)
    gtk_label_set_text (GTK_LABEL (game->p1_score_l), "0");
  if (game->p2_score_l)
    gtk_label_set_text (GTK_LABEL (game->p2_score_l), "0");

  if (game->options.timeout == LIMIT_PERGAME)
    {
      game->movelimit.time_left =
	((double) game->movelimit.time_max / 2) * 60;
      game->movelimit.time_left_o =
	((double) game->movelimit.time_max / 2) * 60;
    }

  timeout_add_movelimit (game);
}

void game_add_to_score(Game *game, gint score)
{
  char buf[202];
	
  switch(game->current_player) {
	case PLAYER1:
    game->p1_score += score;
    sprintf(buf, "%d", game->p1_score);
    gtk_label_set_text(GTK_LABEL(game->p1_score_l), buf);
    break;
  case PLAYER2:
    game->p2_score += score;
    sprintf(buf, "%d", game->p2_score);
    gtk_label_set_text(GTK_LABEL(game->p2_score_l), buf);
    break;
  default:
		break;
  };
}

void
game_get_names (Game * game, char *p1, char *p2)
{
  switch (game->who_am_i)
    {
    case PLAYER1:
      {
	if (game->mode == GAME_OBSERVE)
	  {
	    strncpy (p1, game->opponent, MAX_PLAYER_NAME);
	    strncpy (p2, game->observed, MAX_PLAYER_NAME);
	  }
	else
	  {
	    strncpy (p1, game->name, MAX_PLAYER_NAME);
	    strncpy (p2, game->opponent, MAX_PLAYER_NAME);
	  }
      }
      break;
    case PLAYER2:
      {
	if (game->mode == GAME_OBSERVE)
	  {
	    strncpy (p1, game->observed, MAX_PLAYER_NAME);
	    strncpy (p2, game->opponent, MAX_PLAYER_NAME);
	  }
	else
	  {
	    strncpy (p1, game->opponent, MAX_PLAYER_NAME);
	    strncpy (p2, game->name, MAX_PLAYER_NAME);
	  }
      }
      break;
    case NONE:			/* this should not occur! */
    default:
      break;
    }
}


void game_start_playing(Game *game, enum gamemode mode)
{
  Player *player;
  char buffer[512];
  char *opponent;
  char _p1_n[MAX_PLAYER_NAME+1];
  char _p2_n[MAX_PLAYER_NAME+1];
  gint num;
  guint mask;

  GtkWidget *table, *label, *win;
  GdkPixmap *s;
  GdkBitmap *b;
	
  game->mode = mode;

  if (game->mode != GAME_OBSERVE)
    game_reset_values(game);
	
  /* this is reset after the again_ack is recvd! */
  game->game_over = FALSE;  
  
  game->vbox = gtk_vbox_new(0, 0);
  gtk_container_set_border_width(GTK_CONTAINER(game->vbox), 2);
  
  table = gtk_table_new(10,12,FALSE);
  gtk_box_pack_start(GTK_BOX(game->vbox), table, TRUE, TRUE, 0);
  
  game->canvas = gtk_drawing_area_new();
  gtk_table_attach(GTK_TABLE(table), game->canvas, 0, 8, 0, 10, 0, 0, 2, 0);
	
  mask = GDK_EXPOSURE_MASK;
  if (mode == GAME_NETWORK
      || mode == GAME_TOURNAMENT
      || mode == GAME_SOLO
      || mode == GAME_AI)
  {
    mask |= GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK;
  }

  /* +50 for the rulers (25 width / height) */
  gtk_widget_set_usize(game->canvas, 
                       8*50+RULER_OFFSET*2, 
		       8*50+RULER_OFFSET*2);
  gtk_widget_set_events(game->canvas,
                        gtk_widget_get_events(game->canvas) | mask);
  gdk_color_parse("LightCoral", &game->col_red_after);
  gdk_color_parse("firebrick", &game->col_red);
  gdk_color_parse("LightBlue", &game->col_blue_after);
  gdk_color_parse("MidnightBlue", &game->col_blue);
  gdk_color_white(gtk_widget_get_colormap(game->canvas), &game->white);
  gdk_color_black(gtk_widget_get_colormap(game->canvas), &game->black);
  gdk_colormap_alloc_color(gtk_widget_get_colormap(game->canvas), 
                           &game->col_red,  FALSE, FALSE);
  gdk_colormap_alloc_color(gtk_widget_get_colormap(game->canvas), 
                           &game->col_blue, FALSE, FALSE);
  gdk_colormap_alloc_color(gtk_widget_get_colormap(game->canvas), 
                           &game->col_red_after,  FALSE, FALSE);
  gdk_colormap_alloc_color(gtk_widget_get_colormap(game->canvas), 
                           &game->col_blue_after, FALSE, FALSE);
  game->font = gdk_font_load("fixed");
  
  GFXPATH(preferences.theme_data.p1_small,buffer);
  if (gdk_imlib_load_file_to_pixmap(buffer, &s,&b))
    {
      game->small_blue = gtk_pixmap_new(s,b);
      gdk_imlib_free_pixmap(s);
    }
  GFXPATH(preferences.theme_data.p2_small,buffer);
  if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
    {
      game->small_red = gtk_pixmap_new(s, b);
      gdk_imlib_free_pixmap(s);
    }
  gtk_table_attach_defaults(GTK_TABLE(table), game->small_blue, 8, 9, 1, 2);
  gtk_table_attach_defaults(GTK_TABLE(table), game->small_red, 8, 9, 2, 3);

  game_get_names(game, _p1_n, _p2_n);
  game->p1_name = gtk_label_new(_p1_n);
  gtk_label_set_text(GTK_LABEL(game->p1_name), _p1_n);
	
	if (game->current_player == PLAYER1)
		gtk_widget_set_style(game->p1_name, greenstyle);
	else
		gtk_widget_set_style(game->p1_name, blackstyle);
	
  gtk_misc_set_alignment(GTK_MISC(game->p1_name), 0.0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), game->p1_name, 9, 10, 1, 2);

  game->p2_name = gtk_label_new(_p2_n);
  gtk_label_set_text(GTK_LABEL(game->p2_name), _p2_n);
	
	if (game->current_player == PLAYER2)
		gtk_widget_set_style(game->p2_name, greenstyle);
	else
		gtk_widget_set_style(game->p2_name, blackstyle);
	
  gtk_misc_set_alignment(GTK_MISC(game->p2_name), 0.0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), game->p2_name, 9, 10, 2, 3);

	if (game->options.timeout != LIMIT_OFF)
		{
			label = gtk_label_new("Time Remaining");
			gtk_table_attach_defaults(GTK_TABLE(table), label, 10, 11, 0, 1);
		}
	
	game->movelimit.mylabel = gtk_label_new("");
	game->movelimit.opponentlabel = gtk_label_new("");
	
	if (game->who_am_i == PLAYER1)
		label = game->movelimit.mylabel;
	else
		label = game->movelimit.opponentlabel;
	gtk_table_attach_defaults(GTK_TABLE(table), label, 10, 11, 1, 2);
	
	if (game->who_am_i == PLAYER2)
		label = game->movelimit.mylabel;
	else
		label = game->movelimit.opponentlabel;
	gtk_table_attach_defaults(GTK_TABLE(table), label, 10, 11, 2, 3);
	
	timeout_add_movelimit(game);
	
  game->p1_score_l = gtk_label_new("0");
  gtk_misc_set_alignment(GTK_MISC(game->p1_score_l), 1.0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), game->p1_score_l,	11, 12, 1, 2);
	
  game->p2_score_l = gtk_label_new("0");
  gtk_misc_set_alignment(GTK_MISC(game->p2_score_l), 1.0, 0.5);
  gtk_table_attach_defaults(GTK_TABLE(table), game->p2_score_l, 11, 12, 2, 3);
	
  game->text = gtk_text_new(NULL, NULL);
  gtk_widget_set_style(game->text, sq_get_widget_style());

  win = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win), 
    GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(win), game->text);

  gtk_table_attach_defaults(GTK_TABLE(table), win, 8, 12, 3, 9);

  game->chatline = gtk_entry_new_with_max_length(64);
  gtk_table_attach(GTK_TABLE(table), game->chatline, 
      8, 11, 9, 10, GTK_FILL | GTK_EXPAND, 0, 0, 0);

  game->cmd_quit = gtk_button_new_with_label("Leave");
  gtk_table_attach(GTK_TABLE(table), game->cmd_quit,
      9,10, 0, 1, GTK_FILL | GTK_EXPAND, 0, 0, 0);
  
  gtk_signal_connect(GTK_OBJECT(game->canvas), "realize",
										 GTK_SIGNAL_FUNC(cb_canvas_realize), NULL);
  gtk_signal_connect(GTK_OBJECT(game->canvas), "expose_event",
										 GTK_SIGNAL_FUNC(cb_canvas_expose), game);
  gtk_signal_connect(GTK_OBJECT(game->canvas), "button_press_event",
										 GTK_SIGNAL_FUNC(cb_canvas_button_press), game);
  gtk_signal_connect(GTK_OBJECT(game->canvas), "button_release_event",
										 GTK_SIGNAL_FUNC(cb_canvas_button_release), game);
  gtk_signal_connect(GTK_OBJECT(game->chatline), "activate",
										 GTK_SIGNAL_FUNC(cb_chat), game);
  gtk_signal_connect(GTK_OBJECT(game->cmd_quit), "clicked",
										 GTK_SIGNAL_FUNC(cb_leave), game);

  GFXPATH(preferences.theme_data.p2_stone,buffer);
  if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
    {
      game->spot_red = gtk_pixmap_new(s, b);
      gdk_imlib_free_pixmap(s);
    }
  GFXPATH(preferences.theme_data.p1_stone,buffer);
  if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
    {
      game->spot_blue = gtk_pixmap_new(s, b);
      gdk_imlib_free_pixmap(s);
    }
  GFXPATH(preferences.theme_data.p2_last,buffer);
  if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
    {
      game->spot_red_last = gtk_pixmap_new(s, b);
      gdk_imlib_free_pixmap(s);
    }
  GFXPATH(preferences.theme_data.p1_last,buffer);
  if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
    {
      game->spot_blue_last = gtk_pixmap_new(s, b);
      gdk_imlib_free_pixmap(s);
    }
	GFXPATH(preferences.theme_data.marker, buffer);
	if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
		{
			game->bg_marker = gtk_pixmap_new(s, b);
			gdk_imlib_free_pixmap(s);
		}
	GFXPATH(preferences.theme_data.background, buffer);
	if (gdk_imlib_load_file_to_pixmap(buffer, &s, &b))
		{
			game->bg_tile = gtk_pixmap_new(s, b);
			gdk_imlib_free_pixmap(s);
		}
	
  game->tablabel = gtk_label_new("");
  gtk_label_set_text(GTK_LABEL(game->tablabel), game->opponent);
  gtk_widget_show(game->tablabel);

  gtk_widget_show_all(table);
  gtk_widget_show(game->vbox);
  gtk_notebook_append_page(GTK_NOTEBOOK(top_notebook), game->vbox,
      game->tablabel);
  
  while ( gtk_events_pending() )
    gtk_main_iteration();
	
  num = gtk_notebook_page_num(GTK_NOTEBOOK(top_notebook), game->vbox);
  gtk_notebook_set_page(GTK_NOTEBOOK(top_notebook), num);
  board_draw_background(game);
  player = find_player_by_nick(game->connection, preferences.nickname);
  opponent = malloc(strlen(game->opponent)+1);
  strcpy(opponent, game->opponent);

  if (game->mode != GAME_SOLO && game->mode != GAME_AI)
    ADD_LIST_ITEM(player->opponents, opponent);
  else
    free(opponent);

  if (game->mode == GAME_AI)
    ai_is_playing = 1;

  /* associate plugindata with game, as plugin->data gets overwritten soon */
  if ( GAME_PLUGIN(game) )
    game->plugindata = GAME_PLUGIN(game)->data;
  /* time to handle OnGameStart to the plugin */
  if ( GAME_PLUGIN(game) && GAME_PLUGIN(game)->callbacks->OnGameStart )
    GAME_PLUGIN(game)->callbacks->OnGameStart(game);

  if (game->mode == GAME_AI && game->options.first_player != YOU_FIRST)
  {
    fprintf(stderr, "AI begins\n");
    game->current_player = PLAYER1;
    game->who_am_i = PLAYER1;
    game_next_move(game);
  }
	
}

void
game_next_move (Game * game)
{
  char _p1_n[202], _p2_n[202];

  game->current_move++;
  game->made_square = FALSE;
  game_get_names (game, _p1_n, _p2_n);

  if (game->current_player == PLAYER1)
    {
      game->current_player = PLAYER2;
      gtk_widget_set_style (game->p1_name, blackstyle);
      gtk_widget_set_style (game->p2_name, greenstyle);
    }
  else
    {
      game->current_player = PLAYER1;
      gtk_widget_set_style (game->p1_name, greenstyle);
      gtk_widget_set_style (game->p2_name, blackstyle);
    }


  /* check if game is over; this is needed here too, to prevent
   * the AI to be allowed to make another move although the player
   * already won
   */
  /* game is over if anybody got 150 points and it 15+ in the lead */
  game->game_over = IS_GAME_OVER || game->current_move >= 64;

  if (game->tablabel == current_tablabel && game->game_over == FALSE)
    {
      if (game->current_player == game->who_am_i)
	{
	  if (game->mode == GAME_OBSERVE)
	    sq_status_set ("It's your turn, %s", game->opponent);
	  else
	    sq_status_set ("It's your turn, %s", game->name);
	}
      else
	{
	  int move;
	  if (game->mode == GAME_OBSERVE)
	    sq_status_set ("Wait for %s to make a move", game->observed);
	  else
	    sq_status_set ("Wait for %s to make a move", game->opponent);

	  if (game->mode == GAME_AI)
	    {
	      while (gtk_events_pending ())
		gtk_main_iteration ();
	      move = ai_move ();
	      game_canvas_put_stone (game, move % 8, move / 8);
	    }
	}
    }
  /* #####
     if (game->who_am_i == game->current_player)
     timeout_add_movelimit(game);
     else
     timeout_del_movelimit(game);
   */
  if (game->options.timeout == LIMIT_PERMOVE)
    {
      game->movelimit.time_left = game->movelimit.time_max;
      game->movelimit.time_left_o = game->movelimit.time_max;
    }

  while (gtk_events_pending ())
    gtk_main_iteration ();

  game->game_over = IS_GAME_OVER || game->current_move >= 64;
  game->board_is_full = (game->current_move >= 64);

  if (game->game_over)
    game_is_over (game);
}

/* beware: ugly kludge code follows! */
void
game_gameover_get_captions (Game * game, struct gameover_captions *data)
{
  /* did we fail to make a move in time?? */
  if (game->current_player == game->who_am_i && game->movelimit.timed_out)
    {
      /* yes, we timed out */
      sq_status_set ("%s is the winner because you timed out",
		     game->opponent);
      snprintf (data->message, BUFSIZE,
		"Damn! You timed out. %s won!", game->e_opponent->nickname);
      snprintf (data->ok, BUFSIZE, "Revenge!");
      snprintf (data->abort, BUFSIZE, "Aahhh no. Escape!");
    }
  else if (game->current_player != game->who_am_i
	   && game->movelimit.timed_out)
    {
      sq_status_set ("You are the winner because %s timed out",
		     game->opponent);
      snprintf (data->message, BUFSIZE,
		"Boring! %s timed out. You won!", game->opponent);
      snprintf (data->ok, BUFSIZE, "Lemme do it again!");
      snprintf (data->abort, BUFSIZE, "Booooring! I'm outta here");
    }
  else if (game->board_is_full == TRUE && !(IS_GAME_OVER))
    {
      sq_status_set ("Game Over. TIE!");
      snprintf (data->message, BUFSIZE, "Whoa. It's a tie.");
      snprintf (data->ok, BUFSIZE, "Lemme try it again!");
      snprintf (data->abort, BUFSIZE, "Booooooring! I'm outta here");
    }
  else if (game->who_am_i == PLAYER1 && (IS_GAME_OVER)
	   && game->p1_score > game->p2_score)
    {
      sq_status_set ("%s is the winner", game->name);
      snprintf (data->message, BUFSIZE,
		"Great! You kicked %s's butt!", game->opponent);
      snprintf (data->ok, BUFSIZE, "Lemme do it again!");
      snprintf (data->abort, BUFSIZE, "Booooring! I'm outta here");
    }
  else if (game->who_am_i == PLAYER1 && (IS_GAME_OVER)
	   && game->p1_score < game->p2_score)
    {
      sq_status_set ("%s is the winner", game->opponent);
      snprintf (data->message, BUFSIZE,
		"Damn! Your butt's been kicked by %s!", game->opponent);
      snprintf (data->ok, BUFSIZE, "Revenge!");
      snprintf (data->abort, BUFSIZE, "Aahhh no. Escape!");
    }
  else if (game->who_am_i == PLAYER2 && (IS_GAME_OVER)
	   && game->p2_score > game->p1_score)
    {
      sq_status_set ("%s is the winner", game->name);
      snprintf (data->message, BUFSIZE,
		"Great! You kicked %s's butt!", game->opponent);
      snprintf (data->ok, BUFSIZE, "Lemme do it again!");
      snprintf (data->abort, BUFSIZE, "Booooring! I'm outta here");
    }
  else if (game->who_am_i == PLAYER2 && (IS_GAME_OVER)
	   && game->p2_score < game->p1_score)
    {
      sq_status_set ("%s is the winner", game->opponent);
      snprintf (data->message, BUFSIZE,
		"Damn! Your butt's been kicked by %s!", game->opponent);
      snprintf (data->ok, BUFSIZE, "Revenge!");
      snprintf (data->abort, BUFSIZE, "Aahhh no. Escape!");
    }
  else
    {
      snprintf (data->message, BUFSIZE, "Game Over");
      snprintf (data->ok, BUFSIZE, "Lemme try again!");
      snprintf (data->abort, BUFSIZE, "Boooring! Let me out");
      sq_status_set ("ERROR: nobody won");
    }

}

void
game_gameover_dialog (Game * game,
		      const char *message,
		      const char *button_ack_caption,
		      const char *button_nack_caption)
{
  GtkWidget *yes, *no, *label;
  char buffer[BUFSIZE];

  /* necessary to prevent two dialogs from opening */
  if (game->dialog_shown == TRUE)
    return;

  if (GAME_PLUGIN (game) != 0)
    snprintf (buffer, BUFSIZE, "Game Over! (%s)",
	      GAME_PLUGIN (game)->callbacks->name);
  else
    snprintf (buffer, BUFSIZE, "Game Over!");

  game->dialog = gtk_dialog_new ();
  gtk_window_set_position (GTK_WINDOW (game->dialog), GTK_WIN_POS_CENTER);
  gtk_window_set_title (GTK_WINDOW (game->dialog), buffer);
  gtk_container_border_width (GTK_CONTAINER (game->dialog), 10);

  label = sq_label_new ("%s", message);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (game->dialog)->vbox),
		      label, TRUE, TRUE, 0);

  yes = gtk_button_new_with_label (button_ack_caption);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (game->dialog)->action_area),
		      yes, TRUE, TRUE, 0);

  no = gtk_button_new_with_label (button_nack_caption);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (game->dialog)->action_area),
		      no, TRUE, TRUE, 0);

  gtk_signal_connect (GTK_OBJECT (yes), "clicked",
		      GTK_SIGNAL_FUNC (cb_gameover_again), (gpointer) game);
  gtk_signal_connect (GTK_OBJECT (no), "clicked",
		      GTK_SIGNAL_FUNC (cb_gameover_giveup), (gpointer) game);

  gtk_widget_show_all (GTK_DIALOG (game->dialog)->vbox);
  gtk_widget_show_all (GTK_DIALOG (game->dialog)->action_area);
  gtk_widget_show (game->dialog);
  game->dialog_shown = TRUE;
}

void game_is_over(Game *game)
{
	struct gameover_captions captions;
	int plugin_ret = -1;
	
	bzero(&captions, sizeof(struct gameover_captions));
	/* movelimit timeout not needed anymore */
	timeout_del_movelimit(game);      

	if (game->game_over == FALSE || game->dialog_shown == TRUE)
		return;

	/* call the gameover callback, if present */
	if (GAME_PLUGIN(game) != 0 && GAME_PLUGIN(game)->callbacks->OnGameOver != 0)
		plugin_ret = GAME_PLUGIN(game)->callbacks->OnGameOver(game);
	
	if (plugin_ret == PLUGIN_SUCCESS)
		return;
	
	if (game->mode == GAME_OBSERVE)
		{
			game_chatmessage(game, STARTON " This game is over\n");
			return;
		}
	
  if (game->mode != GAME_OBSERVE)
    {
			
      if ( (game->mode == GAME_NETWORK
						|| game->mode == GAME_TOURNAMENT)
					 && CHATROOM(game->connection)->gamehost[0] != 0)
				{
					sqirc_send(game->connection,
										 "PRIVMSG %s :\001GAMEDONE %s %i %i %i -\001\r\n",
										 CHATROOM(game->connection)->gamehost,
										 game->opponent,
										 game->p1_score,
										 game->p2_score,
										 game->who_am_i);
				}
			
			game_gameover_get_captions(game, &captions);
			game_gameover_dialog(game,captions.message,captions.ok,captions.abort);
    }
}  

void game_chatmessage(Game *game, char *fmt, ...)
{
	char data[BUFSIZE*2];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(data, BUFSIZE*2, fmt, ap);
	va_end(ap);
	
	if (game->text)
		PrintText(game->text, data);
	else
		sq_print_on_current_chatwindow("[PRIVMSG] %s", data);
}

void game_notify_quit(Game *game)
{
  char buffer[512];
  GtkWidget *label, *button;
	
  if (game->timeout_play)
		timeout_del_pending_play_request(game);
  if (game->timeout_observe)
		timeout_del_pending_observe_request(game);

  game->playing = 0;
  if (GTK_IS_WIDGET(game->dialog))
    gtk_widget_destroy(game->dialog);

  game->dialog = gtk_dialog_new();
  gtk_window_set_position(GTK_WINDOW(game->dialog), GTK_WIN_POS_CENTER);
  gtk_container_border_width(GTK_CONTAINER(game->dialog), 5);
	
  if (game->mode == GAME_OBSERVE)
    sprintf(buffer, "Player %s from an observed game left", game->opponent);
  else 
    sprintf(buffer, "Your opponent %s gave up", game->opponent);
  label = gtk_label_new(buffer);
  button = gtk_button_new_with_label("what a sissy!");
  
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(game->dialog)->vbox),
										 label, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(game->dialog)->action_area),
										 button, TRUE, TRUE, 0);
  
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
										 GTK_SIGNAL_FUNC(cb_game_notify_quit), (gpointer)game);
  
  gtk_widget_show_all(game->dialog);
  game->dialog_shown = TRUE;
}

/* this functions keeps track which tablabels we have to highlight
 * when we get new net input
 */
void cb_notebook_switch(GtkWidget *w, GtkNotebookPage *p, guint c)
{
  Game *game = 0;
	Chatroom *chatroom = 0;
	
  /* if it's the chatroom-tab, make sure is set correctly */
  /* current_tablabel = chatroom_tablabel; */
  if ( (game = find_game_by_vbox(p->child)) )
    {
      /* and only reset it, if there is some running game */
      current_tablabel = game->tablabel;
      gtk_widget_set_style(game->tablabel, blackstyle);
			
      if (game->current_player == game->who_am_i)
				sq_status_set("It's your turn, %s", game->name);
			else
				sq_status_set("Wait for %s to make a move", game->opponent);
    }
  else if ( (chatroom = find_chatroom_by_table(p->child)) )
		if (blackstyle != 0)
			{
				current_tablabel = chatroom->tablabel;
				gtk_widget_set_style(chatroom->tablabel, blackstyle);
#if DEBUG || NET_DEBUG				
				if (chatroom_debuglabel)
					gtk_widget_set_style(chatroom_debuglabel, blackstyle);
#endif
			}
}

void cb_game_notify_quit(GtkWidget *w, gpointer data)
{
  Game *game = (Game *)data;

  assert(game != 0);
  game_cleanup(game);
}

void cb_leave(GtkWidget *w, gpointer data)
{
  Game *game = (Game *)data;
  
	game_reset_values(game);
  if (game->mode != GAME_SOLO && game->mode != GAME_AI)
    {
			
			if (game->mode == GAME_OBSERVE)
				sqirc_send(game->connection,
									 "NOTICE %s,%s :GAMEQUIT\r\n",
									 game->opponent, game->observed);
			else
				sqirc_send(game->connection,
									 "NOTICE %s :GAMEQUIT\r\n", game->opponent);
			
      if (game->mode != GAME_OBSERVE) 
				{
					/* tell all observers that we are leaving */
					sq_game_observer_notify(game, "NOTICE", "GAMEQUIT");
					
					/* everytime a player gives up, his room-ranking drops by one!!
					 * also inform the game-host that we're giving up
					 */
					if (chatroom_player_inlist(CHATROOM(game->connection),"Squaroid")
							&& (game->mode == GAME_NETWORK
									|| game->mode == GAME_TOURNAMENT))
						sqirc_send(game->connection, "NOTICE Squaroid :GAMEQUIT\r\n");
				}
      else 
				{
					char *opponent = find_player_opponent_by_nick(game->e_me,
																												game->name);
					DEL_LIST_ITEM(game->e_me->opponents, opponent);
				}
    }
	
	if (game->mode == GAME_AI)
		ai_is_playing = 0;
	
  while (gtk_events_pending())
    gtk_main_iteration();
  
  game->game_over = TRUE;
  current_tablabel = 0;
	game_cleanup(game);
}

void
cb_gameover_again (GtkWidget * e, gpointer data)
{
  Game *game = (Game *) data;

  if (!game)
    return;
  if (game->mode == GAME_NETWORK || game->mode == GAME_TOURNAMENT)
    sqirc_send (game->connection,
		"NOTICE %s :INVITE_AGAIN_ACK\r\n", game->opponent);
  else if (game->mode == GAME_SOLO || game->mode == GAME_AI)
    game->game_over = FALSE;

  gtk_widget_destroy (game->dialog);
  game->dialog_shown = FALSE;
  game->dialog = 0;
  game_reset_values (game);
  board_draw_background (game);

  if (game->mode == GAME_AI && game->options.first_player != YOU_FIRST)
    game_next_move(game);

  /* notify observers to reset their boards */
  sq_game_observer_notify (game, "NOTICE", "OBSERVE_RESET");
}

void cb_gameover_giveup(GtkWidget *e, gpointer data)
{
  char *opponent;
  Game *game = (Game *)data;
  Player *player = find_player_by_nick(game->connection, preferences.nickname);
  assert(game != 0);
  
  if (game->mode == GAME_NETWORK || game->mode == GAME_TOURNAMENT)
		sqirc_send(game->connection,
							 "NOTICE %s :INVITE_AGAIN_REJ\r\n", game->opponent);
  if (game->mode != GAME_SOLO && game->mode != GAME_AI)
    {
      opponent = find_player_opponent_by_nick(player, game->opponent);
      DEL_LIST_ITEM(player->opponents, opponent);
      free(opponent);
    }
  
  game_reset_values(game);
  if (game->mode == GAME_AI)
		{
			ai_is_playing = 0;
			game->game_over = FALSE; /* must do this for the game-over dialog
																		 to not popup again! */
		}
  else
		game->game_over = TRUE;
  if (game->dialog && game->dialog_shown)
    gtk_widget_destroy(game->dialog);
  game->dialog = 0;
  game->dialog_shown = FALSE;
  
  current_tablabel = 0;
  game->playing = FALSE;
  game_cleanup(game);
}

void cb_chat(GtkWidget *e, gpointer data)
{
  Game *game = (Game *)data;

  if (!game) return;
  if (game->mode != GAME_SOLO && game->mode != GAME_AI)
		{
			if (game->mode == GAME_OBSERVE)
				sqirc_send(game->connection,
									 "PRIVMSG %s,%s :%s\r\n",
									 game->opponent, game->observed, 
									 gtk_entry_get_text(GTK_ENTRY(game->chatline)));
			else
				{
					sqirc_send(game->connection,
										 "PRIVMSG %s :%s\r\n", game->opponent,
										 gtk_entry_get_text(GTK_ENTRY(game->chatline)));
					sq_game_observer_notify(game, "PRIVMSG",
																	gtk_entry_get_text(GTK_ENTRY(game->chatline)));
				}
		}
	
  game_chatmessage(game, CHATMESSAGE, preferences.nickname,
									 gtk_entry_get_text(GTK_ENTRY(game->chatline)));
	
  gtk_entry_set_text(GTK_ENTRY(game->chatline), "");
}
