/*
 * sound.c - extendible sound system
 *
 * Copyright (c) 2000 Lukas Schroeder <lukas@azzit.de>
 */
#if SOUND

#include <stdio.h>
#ifdef __STDC__
#include <stdlib.h>
#endif
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/wait.h>

#include "sound.h"
#include "config.h"

char           unixSoundPath[256] = DEFAULT_SFXDIR;
char           unixSoundDev[256] = DEFAULT_SFXDEV;

static int      soundfd;
static char     audioOK = 1;
int             child, player;

struct command
{
	char cmd[32];                           /* 'command' of audio file */
	char filename[256];                     /* filename to read on command */
};
static struct command **library = 0;      /* list of known 'commands' */
static unsigned int library_count = 0;    /* # of entries in library */

int get_file_size(const char *filename)
{
	struct stat buf;
	
	stat(filename, &buf);
	return buf.st_size;
}

void library_add_entry(const char *dir, char *filename)
{
	char *cp1, *cp2, *eob;

	if (!audioOK)
		return;

	library = realloc(library, sizeof(void*) * (library_count + 2));
	
	library[library_count] = malloc(sizeof(struct command));
	bzero(library[library_count], sizeof(struct command));
	
	cp1 = filename;
	cp2 = library[library_count]->cmd;
	eob = cp2 + 32 - 1;
	
	bzero(library[library_count]->cmd, 32);
	while ( *cp1 != 0 && *cp1 != '.' && cp2 < eob )
		*cp2++ = *cp1++;
	*cp2 = 0;
	snprintf(library[library_count]->filename, 256, "%s/%s", dir, filename);

	library[++library_count] = 0;
}

void library_play_entry(const char *cmd)
{
	int i = 0, b = 0;
	char buf[1];
	struct timeval tv;

	if (!audioOK)
		return;

	for (i=0; i < library_count && library[i]; i++)
		{
			if (!strncmp(library[i]->cmd, cmd, 32))
				break;
		}

	if (!library[i])                    /* unknown */
		{
			fprintf(stderr, "Unknown entry '%s'\n", cmd); 
			return;
		}
	
	if ( (player = fork()) == 0)
		{
			/* child */
			int fd_audio, fd_dsp;
			
			if (strstr(library[i]->filename, "raw"))
				{
					fd_audio = open(library[i]->filename, O_RDONLY);
					fd_dsp = open("/dev/dsp", O_WRONLY);
					b = 0;
					while ( (read(fd_audio, buf, 1)) > 0 )
						{
							write(fd_dsp, buf, 1);
							b++;
						}
					close(fd_dsp);
					close(fd_audio);
				}
			else if (strstr(library[i]->filename, "wav"))
				wav_play_file(library[i]->filename);
			_exit(1);
		}

	/* sleep for 300 ms so that the player cat at least hear we're here! */
	tv.tv_sec = 0;
	tv.tv_usec = 300000; 
	select(0, NULL, NULL, NULL, &tv);

	/*	fprintf(stderr, "Read %d of %d bytes for %s:%s\n",
			b, get_file_size(library[i]->filename),
			cmd, library[i]->filename); */
}

void library_stop_sound()
{
	if (player > 0)
		{
			kill(player, SIGTERM);
			player = -1;
		}
}

void sig_chld()
{
	pid_t pid;
	int stat;
        
	while ( (pid = waitpid(-1, &stat, WNOHANG)) > 0 )
		{
			if (player > 0 && pid == player)  /* clear player, if it's dead */
				player = -1;
		}

	return;
}

/* returns 0 on success
	 and a file descriptor to write to the child in wfd */
int sq_init_sound(const char *path, const char *theme, int *wfd)
{
  int i, fd[2];
	DIR *dir;
	struct dirent *entry;
	fd_set rset;
	char buffer[256];
	
	signal (SIGCHLD, sig_chld);
  signal (SIGPIPE, SIG_IGN);

	if (!theme)
		{
			fprintf(stderr, "no sound theme specified!\n");
			return -1;
		}
	snprintf(unixSoundPath, 256, "%s/%s", path?path:DEFAULT_SFXDIR, theme);
	
  if (unixSoundPath[0] == '?')
    {
      audioOK = 0;
			*wfd = -1;
      return -1;
    };
	
	dir = opendir(unixSoundPath);
	if (!dir)
		{
			audioOK = 0;
			*wfd = -1;
			return -1;
		}
	entry = malloc(sizeof(struct dirent));
	while ( (entry = readdir(dir)) )
		{
			if (entry->d_name[0] == '.')
				continue;
			if (strstr(entry->d_name, ".raw") == 0
					&& strstr(entry->d_name, ".wav") == 0)
				continue;
			
			library_add_entry(unixSoundPath, entry->d_name);
		}
	closedir(dir);
	free(entry);
	
  /*  Create a pipe, set the write end to close when we exec the sound server,
			and set both (is the write end necessary?) ends to non-blocking   */
  pipe (fd);
  soundfd = fd[1];
	
  if ( (child = fork ()) > 0)
    {
			close (fd[0]);
			*wfd = fd[1];
			
			return 0;
		}
	
	close (fd[1]);
	/*	library_play_entry("putstone"); */
	
	for (;;)
		{
			
			FD_ZERO(&rset);
			FD_SET(fd[0], &rset);
			
			select(fd[0]+1, &rset, NULL, NULL, NULL);
			
			if (FD_ISSET(fd[0], &rset))
				{
					int b = 0;
					char ch[1];

					bzero(buffer, 256);
					while ( (read(fd[0], ch, 1)) > 0)
						{
							if (ch[0] == '\n')
								break;
							buffer[b++] = ch[0];
						}

					/*					fprintf(stderr, "got: %s\n", buffer); */
					if (!strncasecmp("play ", buffer, 5))
						library_play_entry(buffer + 5);
					else if (!strncasecmp("stop", buffer, 4))
						library_stop_sound();
					else if (!strncasecmp("quit", buffer, 4))
						break;
					
				}
			
		}
	
	if (library)
		{
			for (i=0;i<library_count;i++)
				free(library[i]);
			free(library);
		}
	_exit(1);
	return 0;
}

#endif /* SOUND */
