#ifndef SQ_HAVE_SQUAROID
#define SQ_HAVE_SQUAROID

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <sys/utsname.h>

#include <glib.h>
#include <gdk_imlib.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "../libsqirc/sqirc.h"             /* libSQIRC stuff */
#include "enums.h"

#define GFXPATH(_FILE, _TARGET) sq_get_gfx_path(_FILE, _TARGET)
#define N_(String) (String)

#define DEFAULT_GFXTHEME          "smooth"
#define DEFAULT_SFXTHEME          "default"
#define DEFAULT_PORT              6667
#define MAX_MESSAGE_SIZE          2050
#define MAX_PLAYER_NAME           202
#define BUFSIZE                   512
#define SMALLBUF                  32

#define SND_PUTSTONE              "putstone"
#define SND_SQUARE                "square"

#define ADD_LIST_ITEM(list, item)    list = g_slist_append(list, item)
#define DEL_LIST_ITEM(list, item)    list = g_slist_remove(list, item)
#define GAME_PLUGIN(_game)           ((Plugin *)_game->options.plugin)
#define CHATROOM(_conn)              ((Chatroom *)_conn->data)
#define CHATROOM_DUMMY               ((void*)0x0000)
#define CHATROOM_UNUSED              0    /* marked for deletion */
#define CHATROOM_MAIN                1    /* associated to a connection */
#define CHATROOM_SUBROOM             2    /* associated to a CHATROOM_MAIN */

#define CHATMESSAGE                  "\0032<\017%s\0032>\003 %s\003\n"
#define PRIVMESSAGE                  "\0033>\017%s\0033<\003 %s\003\n"
#define ACTIONMESSAGE                "\0036*\003 %s %s\003\n"
#define STARTON                      "\0034*\003*\0034*\017"
#define STARTOFF                     "\0033*\003*\0033*\017"
#define DEBUGMESSAGE                 "%s"

/* player flags */
#define PLAYER_HAVE_VERSION          1<<0      /* have received version info */
#define PLAYER_HAVE_PLUGINS          1<<1      /* have received plugin info */

/* first player options */
#define YOU_FIRST                    1
#define OPPONENT_FIRST               2
#define RANDOM_FIRST                 3

/* timeout options */
#define LIMIT_OFF                    1
#define LIMIT_PERMOVE                2
#define LIMIT_PERGAME                3

/* game definitions */
#define NEW_SQUARE                   0
#define OLD_SQUARE                   1

#define ADD_STONE                    0
#define DEL_STONE                    1

#define RULER_OFFSET                 25
#define BGTILE_SIZE             preferences.theme_data.background_tile_size
#define BGTILE_XOFFSET          preferences.theme_data.background_line_xoffset
#define BGTILE_YOFFSET          preferences.theme_data.background_line_yoffset
#define ST_XOFFSET              preferences.theme_data.stone_xoffset
#define ST_YOFFSET              preferences.theme_data.stone_yoffset

struct point
{
  unsigned char x, y;
};

struct movedata
{
  struct point c;
  enum _player owner;
};

struct gameoptions
{
	guint timeout;                       /* LIMIT_OFF, LIMIT_PERGAME, ... */
  guint timeout_seconds;               /* default: 0 (no timeout)  */
  enum _timeout_action timeout_action; /* default: lost a move     */
  
  guint score_limit;                   /* default 150              */
  guint move_limit;                    /* default 0: none          */
  
  gint first_player;                   /* default: YOU_FIRST */
	void *plugin;                        /* plugin to handle the game */
};

struct square
{
  struct point corners[4];
};

typedef struct chatroom
{
	GtkWidget *table;                /* GtkTable for layouting */
	GtkWidget *text;                 /* GtkText for all messages */
	GtkWidget *playerlist;           /* GtkCList for list of players */
	GtkWidget *chatline;             /* GtkEntry to enter a message */
	GtkWidget *tablabel;             /* Tablabel for Chatroom */
	GtkWidget *cmdsay;               /* Say! button */
	
  GtkStyle *text_style;
	
	struct connection *connection;
	GSList *games;
	GSList *players;
	GSList *rooms;                   /* sub-rooms */
	
	int type;                        /* CHATROOM_MAIN, CHATROOM_SUBROOM */
	
	char channel[SMALLBUF];          /* channel name */
	char gamehost[SMALLBUF];         /* nickname of host bot, if any */
	
} Chatroom;

/* for each player in the room we have found out some information;
 * it's stored in an appropriate Player struct
 * this is some sort of cache:  the contents of field last_update
 * are compared on request and if not matching, a full update is done
 */ 
typedef struct
{
  char nickname[64];
	char realname[200];
  gboolean valid;
  unsigned int major_version;
  unsigned int minor_version;
	unsigned int revision;
  time_t last_update;
	gint lag;                        /* to adjust for lag */
	guint score;
	
	struct connection *connection;
  GSList *opponents;
	GSList *plugins;                 /* list of loaded plugins */
	
  /* these are flags to control current state */
  int opponents_list:1;            /* are we currently getting an opponents
																			list */
	ulong flags;                     /* to remember status of player */
} Player;

struct _themedata
{
  char invite[128];
  char observe[128];
  char background[128];
	char marker[128];
  char p1_small[128];
  char p1_last[128];
  char p1_stone[128];
  char p2_small[128];
  char p2_last[128];
  char p2_stone[128];
  char xruler[128];
  char yruler[128];
  int background_line_xoffset;
  int background_line_yoffset;
  int background_tile_size;
  int stone_xoffset;
  int stone_yoffset;
};

struct preferences
{
  struct _themedata theme_data;
  char username[64];
  char nickname[64];
  char gfxtheme[64];
  char gfxdir[256];
  char sfxtheme[64];
  char sfxdir[256];
  char hostname[256];
	char playerid[33];
  unsigned int port;
  unsigned int ignore_motd;
  unsigned int saveonexit;
  unsigned int auto_decline;
  unsigned int observe_decline;
	unsigned int sound;
};

typedef struct move_limit
{
	gint timeout;                  /* returned by gtk_timeout_add */
	gint time_max;                 /* init value for time_left */
	gint time_left;                /* seconds left to make a move */
	gint time_left_o;              /* opponent's remaining time */
	gboolean timed_out;            /* TRUE if we received or sent out TIMEDOUT */
	GtkWidget *mylabel;            /* remaining for me */
	GtkWidget *opponentlabel;      /* remaining for opponent */
} MoveLimit;

/* for each opponent you play against, one instance of struct game (Game)
 * is created and filled
 */
typedef struct
{
	GtkWidget *unit_label;
	GtkWidget *hscale;
	GtkAdjustment *adjustment;
} InvitationData;

typedef struct game
{
  GtkWidget *vbox;
  GtkWidget *canvas;
  GtkWidget *text;
  GtkWidget *chatline;
  GtkWidget *dialog;
  GtkWidget *cmd_quit;
  GtkWidget *tablabel;
  GtkWidget *handle_box;         /* currently unused! */
  GtkWidget *invite_radio;
	InvitationData *invite_data;   /* f.i. the scale of the invitation dialog */
	
  GtkWidget *small_red, *small_blue;
  GtkWidget *spot_red, *spot_red_last;
  GtkWidget *spot_blue, *spot_blue_last;
	GtkWidget *bg_marker;
  GtkWidget *bg_tile;
  GtkWidget *xruler, *yruler;
  GtkWidget *p1_name, *p2_name;
  GtkWidget *p1_score_l, *p2_score_l;
	
  GdkPixmap *buffer;
  GdkColor col_red;
  GdkColor col_blue;
  GdkColor col_red_after;
  GdkColor col_blue_after;
  GdkColor white;
  GdkColor black;
  GdkFont *font;
  GdkGC *gc;
	
  GSList *old_lines;
  GSList *old_stones;
  GSList *observer;
  GSList *memory;                 /* game_alloc allocated memory */

  /*
	 * timeout_* keeps the id of the pending timeout, when the user
	 * requested to play or observe and a CTCP VERSION had to be sent
	 * out and received in advance to check whether the squaroid version
	 * on the other side is valid to understand
   * what we want
   */
  guint timeout_play;
  guint timeout_observe;
	
	MoveLimit movelimit;
	
  gboolean game_over;
	gboolean board_is_full;        /* true if the board is full, but no player
																		won by points */
  gboolean playing;
  gboolean dialog_shown;
	gboolean made_square;
	enum invitation invited;  
  enum _player who_am_i;
  enum _player current_player;
  enum _player points[8][8];
	unsigned int marker[8][8];
  enum gamemode mode;
	ulong style;                   /* game style: the plugin id */
  gint current_move;
	gint p1_score;
	gint p2_score;
  struct movedata moves[65];
  struct movedata point_pressed; /* type movedata to be easily used
																	* with game_point_stone
																	* point_pressed stores the location where
																	* the mouse button was pressed, but not
																	* (yet) released again we might paint
																	* something to that location, and/or
																	* remove it
																	*/
	struct movedata marker_pressed;
	
	//	PluginCallbacks *plugin;       /* the plugin that handles this game */
	struct connection *connection;
  struct gameoptions options;
  Player *e_opponent;
  Player *e_me;
	void *plugindata;               /* slot for plugins to store stuff in,
																		 this is to be free()'d in plugins
																		 OnCleanup!
																	*/
  
  char name[BUFSIZE];
  char opponent[BUFSIZE];
  char observed[BUFSIZE];
  time_t updated;                 /* when was data changed that has public
																		 relevance? */
	
  /* flags */
  int pending_observe:1;           /* did we request the opponent list
																			to popup the selection dialog? */
  int pending_invite:1;
} Game;

void callback_destroy(GtkWidget *, gpointer);
void show_dialog_connect();

int sq_check_files();
void sq_status_set(char *, ...);
void sq_status_unset();
void sq_get_gfx_path(char *, char *);

Player *sq_player_create(Chatroom *);
void sq_player_cleanup(Player *);
void sq_player_cleanup_all(Chatroom *);
gboolean sq_player_clientinfo(Player *);

void sq_cleanup_simple_glist(GSList *);


void sq_notify_wrong_version(Game *);
void sq_notify_wrong_version_player(Player *);
void sq_game_add_observer(Game *, char *);
void sq_game_del_observer(Game *, char *);
void sq_game_all_del_observer(struct connection *, char *);
void sq_game_player_left(struct connection *, char *, char *);
void sq_game_observer_notify(Game *, char *, char *);
void sq_print_on_current_chatwindow(char *, ...);
void sq_connection_cleanup(struct connection *);
GtkStyle *sq_get_widget_style();

int  sq_init_sound (const char *, const char *, int *);
void sq_initialize_soundserver();
void sq_play_sound(const char *);
void sq_play_square_sound(int);
void sq_stop_sound();
void sq_finalize_soundserver();
void sq_notebook_append_page(GtkWidget *, GtkWidget *);

GtkWidget *sq_label_new(char *, ...);
void sq_label_set_text(GtkLabel *, char *, ...);
void sq_entry_set_text(GtkEntry *, char *, ...);

Player *find_player_by_nick(struct connection *, const char *);
char *find_player_opponent_by_nick(Player *player, char *nick);
gboolean sq_player_has_plugin(Player *, const char *);

Game *find_game_by_opponent(struct connection *, const char *);
Game *find_game_by_observed(struct connection *, const char *);
Game *find_game_by_canvas(GtkWidget *);
Game *find_game_by_tablabel(GtkWidget *);
Game *find_game_by_vbox(GtkWidget *);
Game *find_game_by_dialog(GtkWidget *);
Game *find_game_by_observer(struct connection *, char *);
char *find_game_observer_by_nick(Game *, char *);

Chatroom *find_chatroom_by_table(GtkWidget *);
Chatroom *find_chatroom_by_tablabel(GtkWidget *);
Chatroom *find_chatroom_by_connection(struct connection *);
Chatroom *find_chatroom_by_channel(struct connection *, const char *);
Chatroom *find_unconnected_chatroom();

void gtk_clist_clear_pixmap(GtkCList *, int, int);

#ifdef MEMORY_DEBUG
void *squaroid_malloc (int size);
void squaroid_free (void *buf);

#define malloc squaroid_malloc
#define free squaroid_free
#endif

/* and now some stuff we need globally: for more info see globals.c */
extern struct preferences preferences;
extern GdkColor colors[];

extern GSList *chatroom_list;

extern GtkStyle *greenstyle;
extern GtkStyle *blackstyle;

extern GtkWidget *top_window;
extern GtkWidget *top_menubar;
extern GtkWidget *top_notebook;
extern GtkWidget *top_statusbar;
extern GtkWidget *current_tablabel;

#if DEBUG || NET_DEBUG
extern GtkWidget *chatroom_debuglabel;
#endif


#include "net_out.h"
#include "chatroom.h"
#include "cfgfiles.h"
#include "netclient.h"
#include "game.h"
#include "timeouts.h"
#include "types.h"
#include "menu.h"
#include "dirbrowser.h"
#include "plugin.h"

#endif  /* SQ_HAVE_SQUAROID */
